#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module to define background workers
"""

from __future__ import print_function, division, absolute_import

import uuid
from threading import Lock, Condition

from Qt.QtCore import *


class Worker(QThread, object):
    workCompleted = Signal(str, object)
    workFailure = Signal(str, str)

    def __init__(self, app, parent=None):
        super(Worker, self).__init__(parent=parent)

        self._execute_tasks = True
        self._app = app

        self._queue_mutex = Lock()

        self._queue = list()
        self._receivers = dict()

        self._wait_condition = Condition(self._queue_mutex)

    def stop(self, wait_for_completion=True):
        """
        Stops the worker, run this before shutdown
        """

        with self._queue_mutex:
            self._execute_tasks = False
            self._wait_condition.notify_all()

        if wait_for_completion:
            self.wait()

    def clear(self):
        """
        Empties the queue
        """

        with self._queue_mutex:
            self._queue = list()

    def queue_work(self, worker_fn, params, asap=False):
        """
        Queues up some work returning a unique id to identify this worker
        :param worker_fn:
        :param params:
        :param asap:
        :return:
        """

        uid = uuid.uuid4().hex
        work = {'id': uid, 'fn': worker_fn, 'params': params}
        with self._queue_mutex:
            if asap:
                self._queue.insert(0, work)
            else:
                self._queue.append(work)

            self._wait_condition.notify_all()

        return uid

    def run(self):
        while self._execute_tasks:
            with self._queue_mutex:
                if len(self._queue) == 0:
                    self._wait_condition.wait()
                    if len(self._queue) == 0:
                        continue
                item_to_process = self._queue.pop()

            if not self._execute_tasks:
                break

            data = None
            try:
                data = item_to_process['fn'](item_to_process['params'])
            except Exception as e:
                if self._execute_tasks:
                    self.workFailure.emit(item_to_process['id'], 'An error ocurred: {}'.format(str(e)))
                else:
                    self.workCompleted.emit(item_to_process['id'], data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collapsible accordion widget similar to Maya Attribute Editor
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc


class AccordionStyle(object):
    BOXED = 1
    ROUNDED = 2
    SQUARE = 3
    MAYA = 4


class AccordionDragDrop(object):
    NO_DRAG_DROP = 0
    INTERNAL_MOVE = 1


class AccordionItem(QGroupBox, object):

    trigger = Signal(bool)

    def __init__(self, accordion, title, widget):
        super(AccordionItem, self).__init__(parent=accordion)

        self._accordion_widget = accordion
        self._widget = widget
        self._rollout_style = AccordionStyle.ROUNDED
        self._drag_drop_mode = AccordionDragDrop.NO_DRAG_DROP
        self._collapsed = False
        self._collapsible = True
        self._clicked = False
        self._custom_data = dict()

        layout = QVBoxLayout()
        layout.setContentsMargins(6, 12, 6, 6)
        layout.setSpacing(0)
        layout.addWidget(widget)

        self.setAcceptDrops(True)
        self.setLayout(layout)
        self.setTitle(title)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_show_menu)

    def get_accordion_widget(self):
        return self._accordion_widget

    def get_widget(self):
        return self._widget

    def get_rollout_style(self):
        return self._rollout_style

    def set_rollout_style(self, style):
        self._rollout_style = style

    def get_drag_drop_mode(self):
        return self._drag_drop_mode

    def set_drag_drop_mode(self, mode):
        self._drag_drop_mode = mode

    def get_collapsible(self):
        return self._collapsible

    def set_collapsible(self, collapsible):
        self._collapsible = collapsible

    accordion_widget = property(get_accordion_widget)
    widget = property(get_widget)
    rollout_style = property(get_rollout_style, set_rollout_style)
    drag_drop_mode = property(get_drag_drop_mode, set_drag_drop_mode)
    collapsible = property(get_collapsible, set_collapsible)

    def enterEvent(self, event):
        self.accordion_widget.leaveEvent(event)
        event.accept()

    def leaveEvent(self, event):
        self.accordion_widget.enterEvent(event)
        event.accept()

    def dragEnterEvent(self, event):
        if not self._drag_drop_mode:
            return
        source = event.source()
        if source != self and source.parent() == self.parent() and isinstance(source, AccordionItem):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        if not self._drag_drop_mode:
            return
        source = event.source()
        if source != self and source.parent() == self.parent() and isinstance(source, AccordionItem):
            event.acceptProposedAction()

    def dropEvent(self, event):
        widget = event.source()
        layout = self.parent().layout()
        layout.insertWidget(layout.indexOf(self), widget)
        self._accordion_widget.emit_items_reordered()

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self.get_drag_drop_rect().contains(event.pos()):
            pixmap = QPixmap.grabWidget(self, self.rect())
            mime_data = QMimeData()
            mime_data.setText('ItemTitle::{}'.format(self.title()))
            drag = QDrag(self)
            drag.setMimeData(mime_data)
            drag.setPixmap(pixmap)
            drag.setHotSpot(event.pos())
            if not drag.exec_():
                self._accordion_widget.emit_item_drag_failed(self)
            event.accept()
        # Check if the expand/collapse should happen
        elif event.button() == Qt.LeftButton and self.expand_collapse_rect().contains(event.pos()):
            self._clicked = True
            event.accept()
        else:
            event.ignore()

    def mouseMoveEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        if self._clicked and self.expand_collapse_rect().contains(event.pos()):
            self.toggle_collapsed()
            event.accept()
        else:
            event.ignore()
        self.clicked = False

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(painter.Antialiasing)
        font = painter.font()
        font.setBold(True)
        painter.setFont(font)

        x = self.rect().x()
        y = self.rect().y()
        w = self.rect().width() - 1
        h = self.rect().height() - 1
        _rect = 8

        if self.rollout_style == AccordionStyle.ROUNDED:
            painter.drawText(x + 33, y + 3, w, 16, Qt.AlignLeft | Qt.AlignTop, self.title())
            self._draw_triangle(painter, x, y)
            pen = QPen(self.palette().color(QPalette.Light))
            pen.setWidthF(0.6)
            painter.setPen(pen)
            painter.drawRoundedRect(x + 1, y + 1, w - 1, h - 1, _rect, _rect)
            pen.setColor(self.palette().color(QPalette.Shadow))
            painter.setPen(pen)
            painter.drawRoundedRect(x, y, w - 1, h - 1, _rect, _rect)
        elif self.rollout_style == AccordionStyle.SQUARE:
            painter.drawText(x + 33, y + 3, w, 16, Qt.AlignLeft | Qt.AlignTop, self.title())
            self._draw_triangle(painter, x, y)
            pen = QPen(self.palette().color(QPalette.Light))
            pen.setWidthF(0.6)
            painter.setPen(pen)
            painter.drawRect(x + 1, y + 1, w - 1, h - 1)
            pen.setColor(self.palette().color(QPalette.Shadow))
            painter.setPen(pen)
            painter.drawRect(x, y, w - 1, h - 1)
        elif self.rollout_style == AccordionStyle.MAYA:
            painter.drawText(x + 33, y + 3, w, 16, Qt.AlignLeft | Qt.AlignTop, self.title())
            painter.setRenderHint(QPainter.Antialiasing, False)
            self._draw_triangle(painter, x, y)
            header_height = 20
            header_rect = QRect(x + 1, y + 1, w - 1, header_height)
            header_rect_shadow = QRect(x - 1, y - 1, w + 1, header_height + 2)
            pen = QPen(self.palette().color(QPalette.Light))
            pen.setWidthF(0.4)
            painter.setPen(pen)
            painter.drawRect(header_rect)
            painter.fillRect(header_rect, QColor(255, 255, 255, 18))
            pen.setColor(self.palette().color(QPalette.Dark))
            painter.setPen(pen)
            painter.drawRect(header_rect_shadow)
            if not self.is_collapsed():
                pen = QPen(self.palette().color(QPalette.Dark))
                pen.setWidthF(0.8)
                painter.setPen(pen)
                offset = header_height + 3
                body_rect = QRect(x, y + offset, w, h - offset)
                body_rect_shadow = QRect(x + 1, y + offset, w + 1, h - offset + 1)
                painter.drawRect(body_rect)
                pen.setColor(self.palette().color(QPalette.Light))
                pen.setWidth(0.4)
                painter.setPen(pen)
                painter.drawRect(body_rect_shadow)
        elif self.rollout_style == AccordionStyle.BOXED:
            if self.is_collapsed():
                a_rect = QRect(x + 1, y + 9, w - 1, 4)
                b_rect = QRect(x, y + 8, w - 1, 4)
                text = '+'
            else:
                a_rect = QRect(x + 1, y + 9, w - 1, h - 9)
                b_rect = QRect(x, y + 8, w - 1, h - 9)
                text = '-'

            pen = QPen(self.palette().color(QPalette.Light))
            pen.setWidthF(0.6)
            painter.setPen(pen)
            painter.drawRect(a_rect)
            pen.setColor(self.palette().color(QPalette.Shadow))
            painter.setPen(pen)
            painter.drawRect(b_rect)
            painter.setRenderHint(painter.Antialiasing, False)
            painter.setBrush(self.palette().color(QPalette.Window).darker(120))
            painter.drawRect(x + 10, y + 1, w - 20, 16)
            painter.drawText(x + 16, y + 1, w - 32, 16, Qt.AlignLeft | Qt.AlignVCenter, text)
            painter.drawText(x + 10, y + 1, w - 20, 16, Qt.AlignCenter, self.title())

        if self.drag_drop_mode:
            rect = self.get_drag_drop_rect()
            _layout = rect.left()
            _rect = rect.right()
            center_y = rect.center().y()
            for y in (center_y - 3, center_y, center_y + 3):
                painter.drawLine(_layout, y, _rect, y)

        painter.end()

    def get_custom_data(self, key, default=None):
        """
        Returns a custom pionter to information stored with this item
        :param key: str
        :param default: variant, default value to return if the key was not found
        :return: variant, data
        """
        return self._custom_data.get(str(key), default)

    def set_custom_data(self, key, value):
        """
        Sets a custom pointer to information stored on this item
        :param key: str
        :param value: variant
        """

        self._custom_data[str(key)] = value

    def get_drag_drop_rect(self):
        return QRect(25, 7, 10, 6)

    def expand_collapse_rect(self):
        return QRect(0, 0, self.width(), 20)

    def is_collapsed(self):
        return self._collapsed

    def set_collapsed(self, state=True):
        if self.collapsible:
            accord = self.accordion_widget
            accord.setUpdatesEnabled(True)
            self._collapsed = state
            if state:
                self.setMinimumHeight(22)
                self.setMaximumHeight(22)
                self.widget.setVisible(False)
            else:
                self.setMinimumHeight(0)
                self.setMaximumHeight(1000000)
                self.widget.setVisible(True)

            self._accordion_widget.emit_item_collapsed(self)
            accord.setUpdatesEnabled(True)

    def toggle_collapsed(self):
        collapsed_state = not self.is_collapsed()
        self.set_collapsed(collapsed_state)
        return collapsed_state

    def _on_show_menu(self):
        if QRect(0, 0, self.width(), 20).contains(self.mapFromGlobal(QCursor.pos())):
            self._accordion_widget.emit_item_menu_requested(self)

    def _draw_triangle(self, painter, x, y):
        brush = QBrush(QColor(255, 255, 255, 160), Qt.SolidPattern)
        if not self.is_collapsed():
            tl, tr, tp = QPoint(x + 9, y + 8), QPoint(x + 19, y + 8), QPoint(x + 14, y + 13.0)
            points = [tl, tr, tp]
            triangle = QPolygon(points)
        else:
            tl, tr, tp = QPoint(x + 11, y + 6), QPoint(x + 16, y + 11), QPoint(x + 11, y + 16.0)
            points = [tl, tr, tp]
            triangle = QPolygon(points)

        current_brush = painter.brush()
        painter.setBrush(brush)
        painter.drawPolygon(triangle)
        painter.setBrush(current_brush)


class AccordionWidget(QScrollArea, object):

    itemCollapsed = Signal(AccordionItem)
    itemMenuRequested = Signal(AccordionItem)
    itemDragFailed = Signal(AccordionItem)
    itemsReordered = Signal()

    def __init__(self, parent=None):
        super(AccordionWidget, self).__init__(parent=parent)

        self._rollout_style = AccordionStyle.SQUARE
        if tpDcc.is_maya():
            self._rollout_style = AccordionStyle.MAYA
        self._drag_drop_mode = AccordionDragDrop.NO_DRAG_DROP
        self._scrolling = False
        self._scroll_init_y = 0
        self._scroll_init_val = 0
        self._item_class = AccordionItem

        self.setFrameShape(QScrollArea.NoFrame)
        self.setAutoFillBackground(False)
        self.setWidgetResizable(True)
        self.setMouseTracking(True)
        self.verticalScrollBar().setMaximumWidth(10)

        widget = QWidget(self)
        layout = QVBoxLayout()
        layout.setContentsMargins(2, 2, 2, 6)
        layout.setSpacing(2)
        layout.addStretch(1)
        widget.setLayout(layout)
        self.setWidget(widget)

    def get_drag_drop_mode(self):
        return self._drag_drop_mode

    def set_drag_drop_mode(self, mode):
        self._drag_drop_mode = mode
        for item in self.findChildren(AccordionItem):
            item.drag_drop_mode = self._drag_drop_mode

    def get_rollout_style(self):
        return self._rollout_style

    def set_rollout_style(self, style):
        self._rollout_style = style
        for item in self.findChildren(AccordionItem):
            item.rollout_style = self._rollout_style

    def get_item_class(self):
        return self._item_class

    def set_item_class(self, item_class):
        self._item_class = item_class

    drag_drop_mode = property(get_drag_drop_mode, set_drag_drop_mode)
    rollout_style = property(get_rollout_style, set_rollout_style)
    item_class = property(get_item_class, set_item_class)

    def eventFilter(self, object, event):
        if event.type() == QEvent.MouseButtonPress:
            self.mousePressEvent(event)
            return True
        elif event.type() == QEvent.MouseMove:
            self.mouseMoveEvent(event)
            return True
        elif event.type() == QEvent.MouseButtonRelease:
            self.mouseReleaseEvent(event)
            return True
        return False

    def enterEvent(self, event):
        if self.can_scroll():
            QApplication.setOverrideCursor(Qt.OpenHandCursor)

    def leaveEvent(self, event):
        if self.can_scroll():
            QApplication.restoreOverrideCursor()

    def mouseMoveEvent(self, event):
        if self._scrolling:
            sbar = self.verticalScrollBar()
            smax = sbar.maximum()
            # Calculate the distance moved for the mouse point
            dy = event.globalY() - self._scroll_init_y
            # Calculate the percentage that is on the scroll bar
            dval = smax * (dy / float(sbar.height()))
            # Calculate the new value
            sbar.setValue(self._scroll_init_val - dval)
        event.accept()

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self.can_scroll():
            self._scrolling = True
            self._scroll_init_y = event.globalY()
            self._scroll_init_val = self.verticalScrollBar().value()
            QApplication.setOverrideCursor(Qt.ClosedHandCursor)
        event.accept()

    def mouseReleaseEvent(self, event):
        if self._scrolling:
            QApplication.restoreOverrideCursor()
        self._scrolling = False
        self._scroll_init_y = 0
        self._scroll_init_val = 0
        event.accept()

    def can_scroll(self):
        return self.verticalScrollBar().maximum() > 0

    def count(self):
        return self.widget().layout().count() - -1

    def set_spacing(self, space_int):
        self.widget().layout().setSpacing(space_int)

    def add_item(self, title, widget, collapsed=False):
        self.setUpdatesEnabled(False)
        try:
            item = self._item_class(self, title, widget)
            item.rollout_style = self.rollout_style
            item.drag_drop_mode = self.drag_drop_mode
            layout = self.widget().layout()
            layout.insertWidget(layout.count() - 1, item)
            layout.setStretchFactor(item, 0)
            if collapsed:
                item.set_collapsed(state=collapsed)
            self.setUpdatesEnabled(True)
            return item
        except Exception:
            self.setUpdatesEnabled(True)
            return None

    def clear(self):
        self.setUpdatesEnabled(False)
        try:
            layout = self.widget().layout()
            while layout.count() > 1:
                item = layout.itemAt(0)

                # First we remove the item from the layout
                w = item.widget
                layout.removeItem(item)

                # Second, close the widget and delete it
                w.close()
                w.deleteLater()
            self.setUpdatesEnabled(True)
        except Exception:
            self.setUpdatesEnabled(True)

    def index_of(self, widget):
        """
        Searches for widget (without including child layouts) and returns the index of widget
        or -1 if the widget is not found
        :param widget: QWidget
        :return: int
        """

        layout = self.widget().layout()
        for index in range(layout.count()):
            if layout.itemAt(index).widget().widget() == widget:
                return index

        return -1

    def is_box_mode(self):
        return self._rollout_style == AccordionStyle.MAYA

    def set_box_mode(self, state):
        if state:
            self._rollout_style = AccordionStyle.BOXED
        else:
            self._rollout_style = AccordionStyle.ROUNDED

    def item_at(self, index):
        layout = self.widget().layout()
        if 0 <= index < layout.count() - 1:
            return layout.itemAt(index).widget()
        return None

    def move_item_down(self, index):
        layout = self.widget().layout()
        if (layout.count() - 1) > (index + 1):
            widget = layout.takeAt(index).widget()
            layout.insertWidget(index + 1, widget)

    def move_item_up(self, index):
        if index > 0:
            layout = self.widget().layout()
            widget = layout.takeAt(index).widget()
            layout.insertWidget(index - 1, widget)

    def take_at(self, index):
        self.setUpdatesEnabled(False)
        try:
            layout = self.widget().layout()
            widget = None
            if 0 < index < layout.count() - 1:
                item = layout.itemAt(index)
                widget = item.widget
                layout.removeItem(item)
                widget.close()
            self.setUpdatesEnabled(True)
            return widget
        except Exception:
            self.setUpdatesEnabled(True)
        return None

    def widget_at(self, index):
        item = self.item_at(index)
        if item:
            return item.widget
        return None

    def emit_item_collapsed(self, item):
        if not self.signalsBlocked():
            self.itemCollapsed.emit(item)

    def emit_item_menu_requested(self, item):
        if not self.signalsBlocked():
            self.itemMenuRequested.emit(item)

    def emit_item_drag_failed(self, item):
        if not self.signalsBlocked():
            self.itemDragFailed.emit(item)

    def emit_items_reordered(self):
        if not self.signalsBlocked():
            self.itemsReordered.emit()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for custom PySide/PyQt windows
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *


class GifLabel(QLabel, object):
    def __init__(self, gif_file):
        super(GifLabel, self).__init__('Name')

        self._movie = QMovie(gif_file, QByteArray(), self)
        self._movie.setCacheMode(QMovie.CacheAll)
        self._movie.setSpeed(100)
        self.setMovie(self._movie)
        self.setAlignment(Qt.AlignCenter)
        self._movie.start()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different widgets used in libraries
"""

from __future__ import print_function, division, absolute_import

import traceback
from functools import partial
from collections import OrderedDict

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc
from tpDcc.libs import qt
from tpDcc.libs.qt.core import animation, image, icon, qtutils, color, pixmap, statusbar
from tpDcc.libs.qt.widgets import progressbar, toolbar, action, buttons

logger = tpDcc.LogsMgr().get_logger()


class LibraryImageSequenceWidget(QToolButton, object):

    DEFAULT_PLAYHEAD_COLOR = QColor(255, 255, 255, 220)
    DEFAULT_PLAYHEAD_HEIGHT = 4

    DEFAULT_STYLE = ("\n"
                     "    QToolBar{\n"
                     "    border: 0px solid black; \n"
                     "    border-radius:2px;\n"
                     "    background-color: rgb(0,0,0,100);\n"
                     "    }\n"
                     "    QToolButton{\n"
                     "    background-color: transparent;\n"
                     "    }\n"
                     "    ")

    def __init__(self, *args):
        super(LibraryImageSequenceWidget, self).__init__(*args)

        self.setMouseTracking(True)

        self._image_sequence = image.ImageSequence('')
        self._image_sequence.frameChanged.connect(self._on_frame_changed)

        self._toolbar = QToolBar(self)
        self._toolbar.setStyleSheet(self.DEFAULT_STYLE)
        animation.fade_out_widget(self._toolbar, duration=0)

        spacer = QWidget()
        spacer.setMaximumWidth(4)
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self._toolbar.addWidget(spacer)

        spacer1 = QWidget()
        spacer1.setMaximumWidth(4)
        spacer1.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self._first_spacer = self._toolbar.addWidget(spacer1)

        self.set_size(150, 150)

    def addAction(self, path, text, tip, callback):
        """
        Overrides base QToolButton addAction function
        Adds aan action to the toolbar
        :param path: str
        :param text: str
        :param tip: str
        :param callback: fn
        :return: QAction
        """

        action_icon = icon.Icon.state_icon(
            path,
            color='rgb(250,250,250,160)',
            color_active='rgb(250,250,250,250)',
            color_disabled='rgb(0,0,0,20)'
        )
        action = QAction(action_icon, text, self._toolbar)
        action.setToolTip(tip)
        self._toolbar.insertAction(self._first_spacer, action)
        action.triggered.connect(callback)

        return action

    def actions(self):
        """
        Overrides base QToolButton actions function
        Returns all the actions that are a child of the toolbar
        :return: list(QAction)
        """

        actions = list()
        for child in self._toolbar.children():
            if isinstance(child, QAction):
                actions.append(child)

        return actions

    def resizeEvent(self, event):
        """
        Overrides base QToolButton resizeEvent function
        Called when the widget is resized
        :param event: QResizeEvent
        """

        self.update_toolbar()

    def enterEvent(self, event):
        """
        Overrides base QToolButton enterEvent function
        Starts playing the image sequence when the mouse enters the widget
        :param event: QEvent
        """

        self._image_sequence.start()
        animation.fade_in_widget(self._toolbar, duration=300)

    def leaveEvent(self, event):
        """
        Overrides base QToolButton leaveEvent function
        Stops playing the image sequence when the mouse leaves the widget
        :param event: QEvent
        """

        self._image_sequence.pause()
        animation.fade_out_widget(self._toolbar, duration=300)

    def mouseMoveEvent(self, event):
        """
        Overrides base QToolButton mouseMoveEvent function
        :param event: QEvent
        """

        if qtutils.is_control_modifier() and self._image_sequence.frame_count() > 1:
            percent = 1.0 - (float(self.width() - event.pos().x()) / float(self.width()))
            frame = int(self._image_sequence.frame_count() * percent)
            self._image_sequence.jump_to_frame(frame)
            frame_icon = self._image_sequence.current_icon()
            self.setIcon(frame_icon)

    def paintEvent(self, event):
        """
        Overrides base QToolButton paintEvent function
        Triggered on frame changed
        :param event: QEvent
        """

        super(LibraryImageSequenceWidget, self).paintEvent(event)

        painter = QPainter()
        painter.begin(self)
        if self.current_filename() and self._image_sequence.frame_count() > 1:
            r = event.rect()
            playhead_height = self.playhead_height()
            playhead_pos = self._image_sequence.percent() * r.width() - 1
            x = r.x()
            y = self.height() - playhead_height
            painter.seten(Qt.NoPen)
            painter.setBrush(QBrush(self.DEFAULT_PLAYHEAD_COLOR))
            painter.drawRect(x, y, playhead_pos, playhead_height)

        painter.end()

    def set_path(self, path):
        """
        Sets a single frame image sequence
        :param path: str
        """

        self._image_sequence.set_path(path)
        self.update_icon()

    def set_dirname(self, dirname):
        """
        Sets the location of the image sequence
        :param dirname: str
        """

        self._image_sequence.set_dirname(dirname)
        self.update_icon()

    def current_filename(self):
        """
        Returns the current image location
        :return: str
        """

        return self._image_sequence.current_filename()

    def playhead_height(self):
        """
        Returns the height of the playhead
        :return: int
        """

        return self.DEFAULT_PLAYHEAD_COLOR

    def set_size(self, w, h):
        """
        Set the size of the widget and updates icon size at the same time
        :param w: int
        :param h: int
        """

        self._size = QSize(w, h)
        self.setIconSize(self._size)
        self.setFixedSize(self._size)

    def update_icon(self):
        """
        Updates the icon for the current frame
        """

        if self._image_sequence.frames():
            frame_icon = self._image_sequence.current_icon()
            self.setIcon(frame_icon)

    def update_toolbar(self):
        """
        Updates the toolbar size depending on the number of actions
        """

        self._toolbar.setIconSize(QSize(16, 20))
        count = len(self.actions()) - 3
        width = 26 * count
        self._toolbar.setGeometry(0, 0, width, 25)
        x = self.rect().center().x() - (self._toolbar.width() * 0.5)
        y = self.height() - self._toolbar.height() - 12
        self._toolbar.setGeometry(x, y, self._toolbar.width(), self._toolbar.height())

    def _on_frame_changed(self, frame=None):
        """
        Internal callback function triggered when the image sequence changes frame
        :param frame: int or None
        """

        if not qtutils.is_control_modifier():
            frame_icon = self._image_sequence.current_icon()
            self.setIcon(frame_icon)


class LibrarySearchWidget(QLineEdit, object):
    SPACE_OPEARTOR = 'and'
    PLACEHOLDER_TEXT = 'Search'

    searchChanged = Signal()

    def __init__(self, parent=None):
        super(LibrarySearchWidget, self).__init__(parent=parent)

        self._library = None
        self._space_operator = 'and'
        search_icon = tpDcc.ResourcesMgr().icon('search', theme='black')
        self._icon_btn = buttons.IconButton(search_icon, icon_padding=2, parent=self)
        self._icon_btn.clicked.connect(self._on_icon_clicked)
        self.set_icon(search_icon)
        cross_icon = tpDcc.ResourcesMgr().icon('delete', theme='black')
        self._clear_btn = buttons.IconButton(cross_icon, icon_padding=2, parent=self)
        self._clear_btn.setCursor(Qt.ArrowCursor)
        self._clear_btn.setToolTip('Clear all search text')
        self._clear_btn.clicked.connect(self._on_clear_clicked)
        self.setPlaceholderText(self.PLACEHOLDER_TEXT)
        self.textChanged.connect(self._on_text_changed)
        self.update()

        tip = 'Search all current items'
        self.setToolTip(tip)
        self.setStatusTip(tip)

        self.setStyleSheet('border-radius: 10px; padding-left: 2px; padding-right: 2px; border: 2px;')

    def update(self):
        """
        Overrides base QLineEdit update function
        """

        self.update_icon_color()
        self.update_clear_button()

    def resizeEvent(self, event):
        """
        Overrides base QLineEdit resizeEvent function
        :param event: QResizeEvent
        """

        super(LibrarySearchWidget, self).resizeEvent(event)

        self.setTextMargins(self.height(), 0, 0, 0)
        size = QSize(self.height(), self.height())
        self._icon_btn.setIconSize(size)
        self._icon_btn.setFixedSize(size)
        self._clear_btn.setIconSize(size)
        x = self.width() - self.height()
        self._clear_btn.setGeometry(x, 0, self.height(), self.height())

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.clear()
        super(LibrarySearchWidget, self).keyPressEvent(event)

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def settings(self):
        """
        Returns a dictionary of the current widget state
        :return: dict
        """

        settings = {
            'text': self.text(),
            'spaceOperator': self.space_operator()
        }

        return settings

    def set_settings(self, settings):
        """
        Restore the widget state from a settings dictonary
        :param settings: dict
        """

        text = settings.get('text', '')
        self.setText(text)
        space_operator = settings.get('spaceOperator')
        if space_operator:
            self.set_space_operator(space_operator)

    def space_operator(self):
        """
        Returns the space operator for the search widget
        :return: str
        """

        return self._space_operator

    def set_space_operator(self, space_operator):
        """
        Sets the space operator
        :param space_operator: str
        """

        self._space_operator = space_operator
        self.search()

    def update_clear_button(self):
        """
        Updates the clera button depending on the current text
        """

        text = self.text()
        if text:
            self._clear_btn.show()
        else:
            self._clear_btn.hide()

    def search(self):
        """
        Run the search query on the library
        """

        if self.library():
            self.library().add_query(self.query())
            self.library().search()
        else:
            qt.logger.info('No library found for the search widget')

        self.update_clear_button()
        self.searchChanged.emit()

    def query(self):
        """
        Returns the query used for the library
        :return: dict
        """

        text = str(self.text())
        filters = list()
        for filter_ in text.split(' '):
            if filter_.split():
                filters.append(('*', 'contains', filter_))
        unique_name = 'searchWidget' + str(id(self))

        return {'name': unique_name, 'operator': self.space_operator(), 'filters': filters}

    def set_icon(self, icon):
        """
        Sets the icon for the search widget
        :param icon: QIcon
        """

        self._icon_btn.setIcon(icon)

    def set_icon_color(self, color):
        """
        Sets the icon color for the search widget icon
        :param color: QColor
        """

        btn_icon = self._icon_btn.icon()
        btn_icon = icon.Icon(btn_icon)
        btn_icon.set_color(color)
        self._icon_btn.setIcon(btn_icon)

        clear_icon = self._clear_btn.icon()
        clear_icon = icon.Icon(clear_icon)
        clear_icon.set_color(color)
        self._clear_btn.setIcon(clear_icon)

    def update_icon_color(self):
        """
        Updates the icon colors to the current foreground role
        """

        clr = self.palette().color(self.foregroundRole())
        clr = color.Color.from_color(clr)
        self.set_icon_color(clr)

    def _on_icon_clicked(self):
        """
        Internal callback function that is triggered when the user clcks on the search icon
        """

        if not self.hasFocus():
            self.setFocus()

    def _on_clear_clicked(self):
        """
        Internal callback function that is triggered when the user clicks the cross icon
        """

        self.setText('')
        self.setFocus()

    def _on_text_changed(self):
        """
        Internal callback function that is triggered when the text changes
        """

        self.search()


class LibraryStatusWidget(statusbar.StatusWidget, object):
    DEFAULT_DISPLAY_TIME = 10000

    INFO_CSS = ''

    WARNING_CSS = """
        color: rgb(240, 240, 240);
        background-color: rgb(240, 170, 0);
    """

    ERROR_CSS = """
        color: rgb(240, 240, 240);
        background-color: rgb(220, 40, 40);
        selection-color: rgb(220, 40, 40);
        selection-background-color: rgb(240, 240, 240);
    """

    def __init__(self, *args):
        super(LibraryStatusWidget, self).__init__(*args)

        self.setFixedHeight(19)
        self.setMinimumWidth(5)

        self.main_layout.setContentsMargins(1, 0, 0, 0)

        self._progress_bar = progressbar.FrameProgressBar(self)
        self._progress_bar.hide()
        self.main_layout.addWidget(self._progress_bar)

    def show_info_message(self, message, msecs=None):
        """
        Overrides progressbar.ProgressBar show_info_message function
        Set an info message to be displayed in the status widget
        :param message: str
        :param msecs: float
        """

        self.setStyleSheet('')
        super(LibraryStatusWidget, self).show_info_message(message, msecs)

    def show_warning_message(self, message, msecs=None):
        """
        Overrides progressbar.ProgressBar show_warning_message function
        Set a warning message to be displayed in the status widget
        :param message: str
        :param msecs: float
        """

        if self.is_blocking():
            return

        self.setStyleSheet(self.WARNING_CSS)
        super(LibraryStatusWidget, self).show_warning_message(message, msecs)

    def show_error_message(self, message, msecs=None):
        """
        Overrides progressbar.ProgressBar show_error_message function
        Set an error message to be displayed in the status widget
        :param message: str
        :param msecs: float
        """

        self.setStyleSheet(self.ERROR_CSS)
        super(LibraryStatusWidget, self).show_error_message(message, msecs)

    def progress_bar(self):
        """
        Returns the progress bar widget
        :return:  progressbar.ProgressBar
        """

        return self._progress_bar


class LibraryToolbarWidget(toolbar.ToolBar, object):

    def __init__(self, *args):
        super(LibraryToolbarWidget, self).__init__(*args)

        self._dpi = 1

    def dpi(self):
        """
        Returns the zoom multiplier
        :return: float
        """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Set the zoom multiplier
        :param dpi: float
        """

        self._dpi = dpi

    def expand_height(self):
        """
        Overrides base toolbar.Toolbar expand_height function
        Returns the height of menu bar when is expanded
        :return: int
        """

        return int(self._expanded_height * self.dpi())

    def collapse_height(self):
        """
        Overrides base toolbar.Toolbar collapse_height function
        Returns the height of widget when collapsed
        :return: int
        """

        return int(self._collapsed_height * self.dpi())


class LibrarySideBarWidgetItem(QTreeWidgetItem, object):
    def __init__(self, *args):
        super(LibrarySideBarWidgetItem, self).__init__(*args)

        self._path = ''
        self._bold = None
        self._icon_path = None
        self._icon_color = None
        self._text_color = None
        self._expanded_icon_path = None
        self._collapsed_icon_path = None

        self._settings = dict()

    def setSelected(self, select):
        """
        Overrides base LibrarySideBarWidget setSelected function
        :param select: bool
        """

        super(LibrarySideBarWidgetItem, self).setSelected(select)
        if select:
            self.set_expanded_parents(select)

    def path(self):
        """
        Returns item path
        :return: str
        """

        return self._path

    def set_path(self, path):
        """
        Sets the path for the item
        :param path: str
        """

        self._path = path

    def default_icon_path(self):
        """
        Returns the default icon path
        :return: str
        """

        return ''

    def expanded_icon_path(self):
        """
        Returns the icon path to be shown when expanded
        :return: str
        """

        return self._expanded_icon_path or tpDcc.ResourcesMgr().get('icons', 'black', 'open_folder')

    def collapsed_icon_path(self):
        """
        Returns the icon path to be shown when collapsed
        :return: str
        """

        return self._collapsed_icon_path or tpDcc.ResourcesMgr().get('icons', 'black', 'folder')

    def icon_path(self):
        """
        Returns the icon path for the item
        :return: str
        """

        return self._icon_path or self.default_icon_path()

    def set_icon_path(self, path):
        """
        Sets the icon path for the item
        :param path: str
        """

        self._icon_path = path
        self.update_icon()

    def default_icon_color(self):
        """
        Returns the default icon color
        :return: str
        """

        palette = self.treeWidget().palette()
        clr = palette.color(self.treeWidget().foregroundRole())
        clr = color.Color.from_color(clr).to_string()

        return str(clr)

    def icon_color(self):
        """
        Returns the icon color
        :return: variant, QColor or None
        """

        return self._icon_color or self.default_icon_color()

    def set_icon_color(self, icon_color):
        """
        Sets the icon color
        :param icon_color: variant, QColor or str
        """

        if isinstance(icon_color, QColor):
            icon_color = color.Color.from_color(icon_color)
        elif isinstance(icon_color, (str, unicode)):
            icon_color = color.Color.from_string(icon_color)

        self._icon_color = icon_color.to_string()
        self.update_icon()

    def textColor(self):
        """
        Returns the foreground color of the item
        :return: QColor
        """

        clr = self.foreground(0).color()
        return color.Color.from_color(clr)

    def setTextColor(self, text_color):
        """
        Sets the foreground color to the given color
        :param text_color: variant, QColor or str
        """

        if isinstance(text_color, QColor):
            text_color = color.Color.from_color(text_color)
        elif isinstance(text_color, (str, unicode)):
            text_color = color.Color.from_string(text_color)
        self._settings['textColor'] = text_color.to_string()
        brush = QBrush()
        brush.setColor(text_color)
        self.setForeground(0, brush)

    def bold(self):
        """
        Returns whether item text is bold or not
        :return: bool
        """

        return self.font(0).bold()

    def set_bold(self, flag):
        """
        Sets whether item text is bold or not
        :param flag: bool
        """

        if flag:
            self._settings['bold'] = flag

        font = self.font(0)
        font.setBold(flag)
        self.setFont(0, font)

    def update_icon(self):
        """
        Forces the icon to update
        """

        path = self.icon_path()
        if not path:
            if self.isExpanded():
                path = self.expanded_icon_path()
            else:
                path = self.collapsed_icon_path()

        clr = self.icon_color()
        px = pixmap.Pixmap(path)
        px.set_color(clr)

        self.setIcon(0, px)

    def url(self):
        """
        Returns the url path
        :return: QUrl
        """

        return QUrl(self.path())

    def parents(self):
        """
        Returns all item parents
        :return: list(LibrarySidebarWidgetItem)
        """

        parents = list()
        parent = self.parent()
        if parent:
            parents.append(parent)
            while parent.parent():
                parent = parent.parent()
                parents.append(parent)

        return parents

    def set_expanded_parents(self, expanded):
        """
        Sets all the parents of the item to the value of expanded
        :param expanded: bool
        """

        parents = self.parents()
        for parent in parents:
            parent.setExpanded(expanded)

    def update(self):
        """
        Updates item
        """

        self.update_icon()

    def settings(self):
        """
        Returns the current state of the item as a dictionary
        :return: dict
        """

        settings = dict()

        is_selected = self.isSelected()
        if is_selected:
            settings['selected'] = is_selected
        is_expanded = self.isExpanded()
        if is_expanded:
            settings['expanded'] = is_expanded
        icon_path = self.icon_path()
        if icon_path != self.default_icon_path():
            settings['iconPath'] = icon_path
        icon_color = self.icon_color()
        if icon_color != self.default_icon_color():
            settings['iconColor'] = icon_color
        bold = self._settings.get('bold')
        if bold:
            settings['bold'] = bold
        text_color = self._settings.get('textColor')
        if text_color:
            settings['textColor'] = text_color

        return settings

    def set_settings(self, settings):
        """
        Sets the current state of the item from a dictionary
        :param settings: dict
        """

        text = settings.get('text')
        if text:
            self.setText(0, text)
        icon_path = settings.get('iconPath')
        if icon_path:
            self.set_icon_path(icon_path)
        icon_color = settings.get('iconColor')
        if icon_color:
            self.set_icon_color(icon_color)
        is_selected = settings.get('selected')
        if is_selected:
            self.setSelected(is_selected)
        is_expanded = settings.get('expanded')
        if is_expanded and self.childCount() > 0:
            self.setExpanded(is_expanded)
        bold = settings.get('bold')
        if bold:
            self.set_bold(bold)
        text_color = settings.get('textColor')
        if text_color:
            self.setTextColor(text_color)


class LibrarySidebarWidget(QTreeWidget, object):

    itemRenamed = Signal(str, str)
    itemDropped = Signal(object)
    itemSelectionChanged = Signal()

    DEFAULT_SEPARATOR = '/'

    @classmethod
    def paths_to_dict(cls, paths, root='', separator=None):
        """
        Returns the given paths as a nested dict
        paths = ['/test/a', '/test/b']
        Result = {'test' : {'a':{}}, {'b':{}}}
        :param paths: list(str)
        :param root: str
        :param separator: str or None
        :return: dict
        """

        separator = separator or cls.DEFAULT_SEPARATOR
        results = OrderedDict()

        for path in paths:
            p = results
            # This is to add support for grouping by the given root path.
            if root and root in path:
                path = path.replace(root, "")
                p = p.setdefault(root, OrderedDict())
            keys = path.split(separator)[0:]
            for key in keys:
                if key:
                    p = p.setdefault(key, OrderedDict())

        return results

    @classmethod
    def find_root(cls, paths, separator=None):
        """
        Finds the common path for the given paths
        :param paths: list(str)
        :param separator: str
        :return: str
        """

        if not paths:
            return

        path = paths[0]
        result = None
        separator = separator or cls.DEFAULT_SEPARATOR
        tokens = path.split(separator)
        for i, token in enumerate(tokens):
            root = separator.join(tokens[:i + 1])
            match = True
            for path in paths:
                if not path.startswith(root + separator):
                    match = False
                    break
            if not match:
                break
            result = root

        return result

    def __init__(self, *args):
        super(LibrarySidebarWidget, self).__init__(*args)

        self._dpi = 1
        self._items = list()
        self._index = dict()
        self._locked = False
        self._library = None
        self._recursive = True
        self._options = {
            'field': 'path',
            'separator': '/',
            'recursive': True,
            'autoRootPath': True,
            'rootText': 'FOLDERS',
            'sortBy': None,
            'queries': [{'filters': [('type', 'is', 'Folder')]}]
        }

        self.set_dpi(1)

        self.setAcceptDrops(True)
        self.setHeaderHidden(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)

        self.itemExpanded.connect(self.update)
        self.itemCollapsed.connect(self.update)

    def update(self, *args):
        """
        Overrides base QTreeWidet update function
        :param args: list
        """

        for item in self.items():
            item.update()

    def selectionChanged(self, *args):
        """
        Overrides base QTreeWidget selectionChanged function
        :param args: list
        """

        self.search()

    def items(self):
        """
        Returns a list of all the items in the tree widget
        :return: list(QTreeWidgetItem)
        """

        items = self.findItems('*', Qt.MatchWildcard | Qt.MatchRecursive)
        return items

    def settings(self):
        """
        Returns a dictionary of the settings for this widget
        :return: dict
        """

        settings = dict()
        vertical_scrollbar = self.verticalScrollBar()
        horizontal_scrollbar = self.horizontalScrollBar()
        settings['verticalScrollBar'] = {'value': vertical_scrollbar.value()}
        settings['horizontalScrollBar'] = {'value': horizontal_scrollbar.value()}
        for item in self.items():
            item_settings = item.settings()
            if item_settings:
                settings[item.path()] = item.settings()

        return settings

    def set_settings(self, settings):
        """
        Sets the settings for this widget
        :param settings: dict
        """

        for path in sorted(settings.keys()):
            s = settings.get(path, None)
            self.set_path_settings(path, s)
        vertical_scrollbar_settings = settings.get('verticalScrollBar', dict())
        value = vertical_scrollbar_settings.get('value', None)
        if value:
            self.verticalScrollBar().setValue(value)
        horizontal_scrollbar_settings = settings.get('horizontalScrollBar', dict())
        value = horizontal_scrollbar_settings.get('value', None)
        if value:
            self.horizontalScrollBar().setValue(value)
        self.setIndentation(18 * self.dpi())

    def set_path_settings(self, path, settings):
        """
        Set paths settings
        :param path: list(str)
        :param settings: dict
        """

        item = self.item_from_path(path)
        if item and settings:
            item.set_settings(settings)

    def set_data(self, data, root='', split=None):
        """
        Sets the items to the given items
        :param data: list(str)
        :param root: str
        :param split: str
        """

        settings = self.settings()
        self.blockSignals(True)
        try:
            self.clear()
            if not root:
                root = self.find_root(data.keys(), self.separator())
            self.add_paths(data, root=root, split=split)
            self.set_settings(settings)
        except Exception as e:
            qt.logger.error('{} | {}'.format(e, traceback.format_exc()))
        finally:
            self.blockSignals(False)

    def add_paths(self, paths, root='', split=None):
        """
        Set the given items as a flat list
        :param paths: list(str)
        :param root: str or None
        :param split: str or None
        """

        data = self.paths_to_dict(paths, root=root, separator=split)
        self.create_items(data, split=split)
        if isinstance(paths, dict):
            self.set_settings(paths)

    def create_items(self, data, split=None):
        """
        Creates the items from the given data dict
        :param data: dict
        :param split: str or None
        """

        split = split or self.DEFAULT_SEPARATOR
        self._index = dict()
        for key in data:
            path = split.join([key])
            item = LibrarySideBarWidgetItem(self)
            item.setText(0, str(key))
            item.set_path(path)
            self._index[path] = item
            if self.root_text():
                item.setText(0, self.root_text())
                item.set_bold(True)
                item.set_icon_path('none')
                item.setExpanded(True)

            def _recursive(parent, children, split=None):
                for text, val in sorted(children.iteritems()):
                    p = parent.path()
                    p = split.join([p, text])
                    child = LibrarySideBarWidgetItem()
                    child.setText(0, str(text))
                    child.set_path(p)
                    parent.addChild(child)
                    self._index[p] = child
                    _recursive(child, val, split=split)

            _recursive(item, data[key], split=split)

        self.update()

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library
        self._library.dataChanged.connect(self._on_data_changed)
        self._on_data_changed()

    def dpi(self):
        """
        Returns the zoom multiplier
        :return: float
        """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Sets zoom multiplier
        :param dpi: float
        """

        size = 24 * dpi
        self.setIndentation(15 * dpi)
        self.setMinimumWidth(35 * dpi)
        self.setIconSize(QSize(size, size))
        self.setStyleSheet('height: {}'.format(size))

    def is_recursive(self):
        """
        Returns whether the recursive query is enabled or not
        :return: bool
        """

        return self._recursive

    def set_recursive(self, enable):
        """
        Sets whether recursive query is enabled or not
        :param enable: bool
        """

        self._recursive = enable
        self.search()

    def is_locked(self):
        """
        Returns whether widget items are read only or not
        :return: bool
        """

        return self._locked

    def set_locked(self, locked):
        """
        Sets whether widget items are read only or not
        :param locked: bool
        """

        self._locked = locked

    def separator(self):
        """
        Returns the separator used in the fields to seaprate level values
        :return: str
        """

        return self._options.get('separator', self.DEFAULT_SEPARATOR)

    def root_text(self):
        """
        Returns the root text
        :return: str
        """

        return self._options.get('rootText')

    def field(self):
        """
        Returns the field
        :return: str
        """

        return self._options.get('field', '')

    def sort_by(self):
        """
        Returns the sort by field
        :return: str
        """

        return self._options.get('sortBy', [self.field()])

    def select_item(self, item):
        """
        Selects given item
        :param item: QTreeWidgetItem
        """

        self.select_paths([item.path()])

    def selected_item(self):
        """
        Returns the current selected item
        :return: QTreeWidgetItem
        """

        path = self.selected_path()
        return self.item_from_path(path)

    def selected_path(self):
        """
        Returns the current selected path
        :return: str
        """
        paths = self.selected_paths()
        if paths:
            return paths[-1]

    def selected_paths(self):
        """
        Returns the paths that are selected
        :return: list(str)
        """

        paths = list()
        items = self.selectedItems()
        for item in items:
            path = item.path()
            paths.append(path)

        return paths

    def select_path(self, path):
        """
        Selects the given path
        :param path: str
        """

        self.select_paths([path])

    def select_paths(self, paths):
        """
        Select the items with the given paths
        :param paths: list(str)
        """

        paths = self._normalize_paths(paths)
        items = self.items()
        for item in items:
            if item.path() in paths:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def select_url(self, url):
        """
        Select the item with the given url
        :param url: str
        """

        items = self.items()
        for item in items:
            if item.url() == url:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def selected_urls(self):
        """
        Returns the urls for the selected items
        :return: list(str)
        """

        urls = list()
        items = self.selectedItems()
        for item in items:
            urls.append(item.url())

        return urls

    def item_from_url(self, url):
        """
        Returns the item for the given URL
        :param url: QUrl
        :return: QTreeWidgetItem
        """

        for item in self.items():
            if url == item.url():
                return item

    def item_from_path(self, path):
        """
        Returns the item for the given path
        :param path: str
        :return: QTreeWidgetItem
        """

        return self._index.get(path)

    def expanded_items(self):
        """
        Returns all the expanded items
        :return: list(QTreeWidgetItem)
        """

        for item in self.items():
            if self.isItemExpanded(item):
                yield item

    def expanded_paths(self):
        """
        Returns all the expanded paths
        :return: list(QUrl)
        """

        for item in self.expanded_items():
            yield item.url()

    def set_expanded_paths(self, paths):
        """
        Stes the given paths as expanded
        :param paths: list(str)
        """

        for item in self.items():
            if item.url() in paths:
                item.setExpanded(True)

    def search(self):
        """
        Runs library search
        """

        if self.library():
            self.library().add_query(self.query())
            self.library().search()
        else:
            qt.logger.info('No library found for the sidebar widget')

    def query(self):
        """
        Returns the query for the sidebar widget
        :return: dict
        """

        filters = list()

        for path in self.selected_paths():
            if self.is_recursive():
                filter_ = ('folder', 'startswith', path + '/')
                filters.append(filter_)
            filter_ = ('folder', 'is', path)
            filters.append(filter_)
        unique_name = 'sidebar_widget_' + str(id(self))
        return {'name': unique_name, 'operator': 'or', 'filters': filters}

    def clear(self):
        """
        Clears all the items from the tree widget
        """

        self._items = list()
        self._index = dict()
        super(LibrarySidebarWidget, self).clear()

    def _normalize_paths(self, paths):
        """
        Normalizes given paths
        :param paths: list(str)
        :return: list(str)
        """

        return [path.replace('\\', '/') for path in paths]

    def _on_data_changed(self):
        """
        Internal callback function that is triggered when the library data changes
        """

        pass


class SortByMenu(QMenu, object):
    def __init__(self, *args, **kwargs):
        super(SortByMenu, self).__init__(*args, **kwargs)

        self._library = None

    def show(self, point=None):
        """
        Overrides base QMenu show function
        :param point: QPoint
        """

        self.clear()

        sort_by = self.library().sort_by()
        if sort_by:
            current_field = self.library().sort_by()[0].split(':')[0]
            current_order = 'dsc' if 'dsc' in self.library().sort_by()[0] else 'asc'
        else:
            current_field = ''
            current_order = ''

        separator_action = action.SeparatorAction('Sort By', self)
        self.addAction(separator_action)

        fields = self.library().SortFields
        for field in fields:
            field_action = self.addAction(field.title())
            field_action.setCheckable(True)
            field_action.setChecked(bool(current_field == field))
            callback = partial(self.set_sort_by, field, current_order)
            field_action.triggered.connect(callback)

        separator_action = action.SeparatorAction('Sort Order', self)
        self.addAction(separator_action)

        ascending_action = self.addAction('Ascending')
        ascending_action.setCheckable(True)
        ascending_action.setChecked(bool(current_order == 'asc'))
        callback = partial(self.set_sort_by, current_field, 'asc')
        ascending_action.triggered.connect(callback)

        descending_action = self.addAction('Descending')
        descending_action.setCheckable(True)
        descending_action.setChecked(bool(current_order == 'dsc'))
        callback = partial(self.set_sort_by, current_field, 'dsc')
        descending_action.triggered.connect(callback)

        if not point:
            point = QCursor.pos()

        self.exec_(point)

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def set_sort_by(self, sort_name, sort_order):
        """
        Sets the sort by value for the library
        :param sort_name: str
        :param sort_order: str
        """

        if sort_name == 'Custom Order':
            sort_order = 'asc'

        value = sort_name + ':' + sort_order
        self.library().set_sort_by([value])
        self.library().search()


class GroupByMenu(QMenu, object):
    def __init__(self, *args, **kwargs):
        super(GroupByMenu, self).__init__(*args, **kwargs)

        self._library = None

    def show(self, point=None):
        """
        Overrides base QMenu show function
        :param point: QPoint
        """

        self.clear()

        group_by = self.library().group_by()
        if group_by:
            current_field = group_by[0].split(':')[0]
            current_order = 'dsc' if 'dsc' in group_by[0]else 'asc'
        else:
            current_field = ''
            current_order = ''

        separator_action = action.SeparatorAction('Group By', self)
        self.addAction(separator_action)

        none_action = self.addAction('None')
        none_action.setCheckable(True)
        callback = partial(self.set_group_by, None, None)
        none_action.triggered.connect(callback)

        fields = self.library().GroupFields
        for field in fields:
            field_action = self.addAction(field.title())
            field_action.setCheckable(True)
            field_action.setChecked(bool(current_field == field))
            callback = partial(self.set_group_by, field, current_order)
            field_action.triggered.connect(callback)

        separator_action = action.SeparatorAction('Group Order', self)
        self.addAction(separator_action)

        ascending_action = self.addAction('Ascending')
        ascending_action.setCheckable(True)
        ascending_action.setChecked(bool(current_order == 'asc'))
        callback = partial(self.set_group_by, current_field, 'asc')
        ascending_action.triggered.connect(callback)

        descending_action = self.addAction('Descending')
        descending_action.setCheckable(True)
        descending_action.setChecked(bool(current_order == 'dsc'))
        callback = partial(self.set_group_by, current_field, 'dsc')
        descending_action.triggered.connect(callback)

        if not point:
            point = QCursor.pos()

        self.exec_(point)

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def set_group_by(self, group_name, group_order):
        """
        Sets the group by value for the library
        :param group_name: str
        :param group_order: str
        """

        if group_name:
            value = [group_name + ':' + group_order]
        else:
            value = None

        self.library().set_group_by(value)
        self.library().search()


class FilterByMenu(QMenu, object):
    def __init__(self, *args, **kwargs):
        super(FilterByMenu, self).__init__(*args, **kwargs)

        self._facets = list()
        self._library = None
        self._options = {'field': 'type'}
        self._settings = dict()

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library
        library.searchStarted.connect(self._on_search_init)

    def _on_search_init(self):
        pass

    def settings(self):
        """
        Returns the settings for the filter menu
        :return: dict
        """

        return self._settings

    def set_settings(self, settings):
        """
        Set the settings for the filter menu
        :param settings: dict
        """

        self._settings = settings

    def name(self):
        """
        Returns the name of the fulter used by the library
        """

        return self._options.get('field') + 'FilterMenu'

    def set_all_enabled(self, enabled):
        """
        Set all the filters enabled
        :param enabled: bool
        """

        for facet in self._facets:
            self._settings[facet.get('name')] = enabled

    def is_show_all_enabled(self):
        """
        Returns whether all current filters are enabled or not
        :return: bool
        """

        for facet in self._facets:
            if not self._settings.get(facet.get('name'), True):
                return False

        return True

    def is_active(self):
        """
        Returns whether are any filters currently active using the settings
        :return: bool
        """

        settings = self.settings()
        for name in settings:
            if not settings.get(name):
                return True

        return False

    def set_options(self, options):
        """
        Sets the options to be used by the filters menu
        :param options: dict
        """

        self._options = options

    def show(self, point=None):
        """
        Shows the menu options
        :param point: variant, QPoint or None
        """

        self.clear()

        field = self._options.get('field')
        queries = self.library().queries(exclude=self.name())
        self._facets = self.library().distinct(field, queries=queries)
        separator_action = action.SeparatorAction('Show {}'.format(field.title()), self)
        self.addAction(separator_action)
        label_action = action.LabelAction('Show All', self)
        self.addAction(label_action)

        label_action.setEnabled(not self.is_show_all_enabled())
        callback = partial(self._on_show_all_clicked)
        label_action.triggered.connect(callback)

        self.addSeparator()

        for facet in self._facets:
            name = facet.get('name', '')
            checked = self.settings().get(name, True)
            filter_by_action = FilterByAction(self)
            filter_by_action.set_facet(facet)
            filter_by_action.setChecked(checked)
            self.addAction(filter_by_action)
            callback = partial(self._on_action_checked, name, not checked)
            filter_by_action.triggered.connect(callback)

        point = point or QCursor.pos()

        self.exec_(point)

    def _on_search_init(self):
        """
        Internal callback function that is triggered before each search to update the filter menu query
        """

        filters = list()
        settings = self.settings()
        field = self._options.get('field')
        for name in settings:
            checked = settings.get(name, True)
            if not checked:
                filters.append((field, 'not', name))

        query = {
            'name': self.name(),
            'operator': 'and',
            'filters': filters
        }

        self.library().add_query(query)

    def _on_action_checked(self, name, checked):
        """
        Internal callback function triggered when an action has been clicked
        :param name: str
        :param checked: bool
        """

        if qtutils.is_control_modifier():
            self.set_all_enabled(False)
            self._settings[name] = True
        else:
            self._settings[name] = checked

        self.library().search()

    def _on_show_all_clicked(self):
        """
        Internal callback function that is triggered when the user clicks the show all action
        """

        self.set_all_enabled(True)
        self.library().search()


class FilterByAction(QWidgetAction, object):
    def __init__(self, parent=None):
        super(FilterByAction, self).__init__(parent)

        self._facet = None
        self._checked = False

    def setChecked(self, checked):
        """
        Overrides base QWidgetAction setChecked function
        :param checked: bool
        """

        self._checked = checked

    def set_facet(self, facet):
        """
        Sets action facet
        :param facet:
        """

        self._facet = facet

    def createWidget(self, menu):
        """
        Overrides base QWidgetAction createWidget function
        :param menu: QMenu
        :return: QWidget
        """

        widget = QFrame(self.parent())
        widget.setObjectName('filterByAction')
        facet = self._facet
        name = facet.get('name', '')
        count = str(facet.get('count', 0))
        title = name.replace('.', '').title()
        label = QCheckBox(widget)
        label.setAttribute(Qt.WA_TransparentForMouseEvents)
        label.setText(title)
        label.installEventFilter(self)
        label.setChecked(self._checked)
        label2 = QLabel(widget)
        label2.setObjectName('actionCounter')
        label2.setText(count)
        layout = QHBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(label, stretch=1)
        layout.addWidget(label2)
        widget.setLayout(layout)

        label.toggled.connect(self._on_triggered)

        return widget

    def _on_triggered(self, checked=None):
        """
        Triggered when teh checkbox value has changed
        :param checked: bool
        """

        self.triggered.emit()
        self.parent().close()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains mixin decorators for Qt
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import qtutils, theme as core_theme


def theme_mixin(cls):
    """
    Decorator that can be used to allow custom widget classes to retrieve theme info of its main tpDcc window
    :param cls:
    :return: cls
    """

    original_init__ = cls.__init__

    def my__init__(self, *args, **kwargs):
        original_init__(self, *args, **kwargs)
        current_theme = self.theme()
        if current_theme:
            self.setStyleSheet(current_theme.stylesheet())

    def polish(self):
        self.style().polish(self)

    def theme(self):
        found_theme = None
        top_widget = self
        while top_widget.parentWidget():
            top_widget = top_widget.parentWidget()
            if hasattr(top_widget, 'theme') and callable(getattr(top_widget, 'theme')):
                found_theme = top_widget.theme()
                break

        if not found_theme:
            found_theme = core_theme.Theme()

        return found_theme

    def theme_default_size(self):
        theme = self.theme()
        if not theme:
            return core_theme.Theme.DEFAULT_SIZE
        else:
            return theme.default_size

    def accent_color(self):
        theme = self.theme()
        if not theme:
            return core_theme.Theme.Colors.BLUE
        else:
            return theme.accent_color

    setattr(cls, '__init__', my__init__)
    setattr(cls, 'polish', polish)
    setattr(cls, 'theme', theme)
    setattr(cls, 'theme_default_size', theme_default_size)
    setattr(cls, 'accent_color', accent_color)

    return cls


def property_mixin(cls):
    """
    Mixin decorator that runs a function after an object dynamic property value changed
    :param cls:
    :return: cls
    """

    def _new_event(self, event):
        if event.type() == QEvent.DynamicPropertyChange:
            property_name = event.propertyName()
            if hasattr(self, '_set_{}'.format(property_name)):
                callback = getattr(self, '_set_{}'.format(property_name))
                callback(self.property(str(property_name)))
        return super(cls, self).event(event)

    setattr(cls, 'event', _new_event)

    return cls


def cursor_mixin(cls):
    """
    Mixin decorator that changes cursor to Qt.PointingHandCursor when mouse is over an enabled widget and to
    Qt.ForbiddenCursor when mouse is over a disabled widget
    :param cls:
    :return: cls
    """

    old_enter_event = cls.enterEvent
    old_leave_event = cls.leaveEvent

    def _new_enter_event(self, *args, **kwargs):
        old_enter_event(self, *args, **kwargs)
        QApplication.setOverrideCursor(Qt.PointingHandCursor if self.isEnabled() else Qt.ForbiddenCursor)
        return super(cls, self).enterEvent(*args, **kwargs)

    def _new_leave_event(self, *args, **kwargs):
        old_leave_event(self, *args, **kwargs)
        QApplication.restoreOverrideCursor()
        return super(cls, self).leaveEvent(*args, **kwargs)

    setattr(cls, 'enterEvent', _new_enter_event)
    setattr(cls, 'leaveEvent', _new_leave_event)

    return cls


def focus_shadow_mixin(cls):
    """
    Adds a shadow effect for decorated class when widget is focused.
    If the widget is focused, shadow effect is enabled; otherwise shadow effect is disabled.
    :param cls:
    :return: cls
    """

    old_focus_in_event = cls.focusInEvent
    old_focus_out_event = cls.focusOutEvent

    def _new_focus_in_event(self, *args, **kwargs):
        old_focus_in_event(self, *args, **kwargs)
        if not self.graphicsEffect():
            shadow_effect = QGraphicsDropShadowEffect(self)
            object_type = self.property('type')
            theme = getattr(cls, 'theme') if hasattr(cls, 'theme') else None
            if theme:
                color = vars(theme).get('{}_color'.format(object_type or 'primary'))
            else:
                color = Qt.red
            shadow_effect.setColor(color)
            shadow_effect.setOffset(0, 0)
            shadow_effect.setBlurRadius(5)
            shadow_effect.setEnabled(False)
            self.setGraphicsEffect(shadow_effect)
        if self.isEnabled():
            self.graphicsEffect().setEnabled(True)

    def _new_focus_out_event(self, *args, **kwargs):
        old_focus_out_event(self, *args, **kwargs)
        if self.graphicsEffect():
            self.graphicsEffect().setEnabled(False)

    setattr(cls, 'focusInEvent', _new_focus_in_event)
    setattr(cls, 'focusOutEvent', _new_focus_out_event)

    return cls


def hover_shadow_mixin(cls):
    """
    Adds a shadow effect for decorated class when widget is hovered.
    If the widget is hovered, shadow effect is enabled; otherwise shadow effect is disabled.
    :param cls:
    :return: cls
    """

    old_enter_event = cls.enterEvent
    old_leave_event = cls.leaveEvent

    def _new_enter_event(self, *args, **kwargs):
        old_enter_event(self, *args, **kwargs)
        if not self.graphicsEffect():
            shadow_effect = QGraphicsDropShadowEffect(self)
            object_type = self.property('type')
            theme = getattr(cls, 'theme') if hasattr(cls, 'theme') else None
            if theme:
                color = vars(theme).get('{}_color'.format(object_type or 'primary'))
            else:
                color = Qt.red
            shadow_effect.setColor(color)
            shadow_effect.setOffset(0, 0)
            shadow_effect.setBlurRadius(5)
            shadow_effect.setEnabled(False)
            self.setGraphicsEffect(shadow_effect)
        if self.isEnabled():
            self.graphicsEffect().setEnabled(True)

    def _new_leave_event(self, *args, **kwargs):
        old_leave_event(self, *args, **kwargs)
        if self.graphicsEffect():
            self.graphicsEffect().setEnabled(False)

    setattr(cls, 'enterEvent', _new_enter_event)
    setattr(cls, 'leaveEvent', _new_leave_event)

    return cls


def stacked_opacity_animation_mixin(cls):
    """
    Decorators for stacked widget
    When stacked widget index changes, show opacity and position animation for current widget
    :param cls:
    :return:
    """

    if not qtutils.is_stackable(cls):
        return cls

    old_init = cls.__init__

    def _new_init(self, *args, **kwargs):
        old_init(self, *args, **kwargs)
        self._prev_index = 0
        self._to_show_pos_anim = QPropertyAnimation()
        self._to_show_pos_anim.setDuration(400)
        self._to_show_pos_anim.setPropertyName('pos')
        self._to_show_pos_anim.setEndValue(QPoint(0, 0))
        self._to_show_pos_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._to_hide_pos_anim = QPropertyAnimation()
        self._to_hide_pos_anim.setDuration(400)
        self._to_hide_pos_anim.setPropertyName('pos')
        self._to_hide_pos_anim.setEndValue(QPoint(0, 0))
        self._to_hide_pos_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._opacity_effect = QGraphicsOpacityEffect()
        self._opacity_anim = QPropertyAnimation()
        self._opacity_anim.setDuration(400)
        self._opacity_anim.setEasingCurve(QEasingCurve.InCubic)
        self._opacity_anim.setPropertyName('opacity')
        self._opacity_anim.setStartValue(0.0)
        self._opacity_anim.setEndValue(1.0)
        self._opacity_anim.setTargetObject(self._opacity_effect)
        self._opacity_anim.finished.connect(self._on_disable_opacity)
        self.currentChanged.connect(self._on_play_anim)

    def _on_play_anim(self, index):
        current_widget = self.widget(index)
        if self._prev_index < index:
            self._to_show_pos_anim.setStartValue(QPoint(self.width(), 0))
            self._to_show_pos_anim.setTargetObject(current_widget)
            self._to_show_pos_anim.start()
        else:
            self._to_hide_pos_anim.setStartValue(QPoint(-self.width(), 0))
            self._to_hide_pos_anim.setTargetObject(current_widget)
            self._to_hide_pos_anim.start()
        current_widget.setGraphicsEffect(self._opacity_effect)
        current_widget.graphicsEffect().setEnabled(True)
        self._opacity_anim.start()
        self._prev_index = index

    def _on_disable_opacity(self):
        self.currentWidget().graphicsEffect().setEnabled(False)

    setattr(cls, '__init__', _new_init)
    setattr(cls, '_on_play_anim', _on_play_anim)
    setattr(cls, '_on_disable_opacity', _on_disable_opacity)

    return cls


class FieldMixin(object):

    computed_dict = None
    properties_dict = None

    def register_field(self, name, getter=None, setter=None, required=False):
        if self.computed_dict is None:
            self.computed_dict = dict()
        if self.properties_dict is None:
            self.properties_dict = dict()
        if callable(getter):
            value = getter()
            self.computed_dict[name] = {
                'value': value,
                'getter': getter,
                'setter': setter,
                'required': required,
                'bind': []
            }
        else:
            self.properties_dict[name] = {
                'value': getter,
                'required': required,
                'bind': []
            }

    def bind(self, data_name, widget, qt_property, index=None, signal=None, callback=None):
        data_dict = {
            'data_name': data_name,
            'widget': widget,
            'widget_property': qt_property,
            'index': index,
            'callback': callback
        }
        if data_name in self.computed_dict:
            self.computed_dict[data_name]['bind'].append(data_dict)
        else:
            self.properties_dict[data_name]['bind'].append(data_dict)
        if signal:
            getattr(widget, signal).connect(partial(self._slot_changed_from_user, data_dict))

        self._data_update_ui(data_dict)

        return widget

    def fields(self):
        return self.properties_dict.keys() + self.computed_dict.keys()

    def field(self, name):
        if name in self.properties_dict:
            return self.properties_dict[name]['value']
        elif name in self.computed_dict:
            new_value = self.computed_dict[name]['getter']()
            self.computed_dict[name]['value'] = new_value
            return new_value
        else:
            raise KeyError('No filed with name: "{}" found!'.format(name))

    def set_field(self, name, value):
        if name in self.properties_dict:
            self.properties_dict[name]['value'] = value
            self._slot_property_changed(name)
        elif name in self.computed_dict:
            self.computed_dict[name]['value'] = value

    def _data_update_ui(self, data_dict):
        data_name = data_dict.get('data_name')
        widget = data_dict['widget']
        index = data_dict['index']
        widget_property = data_dict['widget_property']
        callback = data_dict['callback']
        value = None
        if index is None:
            value = self.field(data_name)
        elif isinstance(self.field(data_name), dict):
            value = self.field(data_name).get(index)
        elif isinstance(self.field(data_name), list):
            value = self.field(data_name)[index] if index < len(self.field(data_name)) else None
        if widget.metaObject().indexOfProperty(widget_property) > -1 \
                or widget_property in map(str, widget.dynamicPropertyNames()):
            widget.setProperty(widget_property, value)
        else:
            widget.set_field(widget_property, value)
        if callable(callback):
            callback()

    def _slot_property_changed(self, property_name):
        for key, setting_dict in self.properties_dict.items():
            if key == property_name:
                for data_dict in setting_dict['bind']:
                    self._data_update_ui(data_dict)

        for key, setting_dict in self.computed_dict.items():
            for data_dict in setting_dict['bind']:
                self._data_update_ui(data_dict)

    def _slot_changed_from_user(self, data_dict, ui_value):
        self._ui_update_data(data_dict, ui_value)

    def _ui_update_data(self, data_dict, ui_value):
        data_name = data_dict.get('data_name')
        index = data_dict.get('index', None)
        if index is None:
            self.set_field(data_name, ui_value)
        else:
            old_value = self.field(data_name)
            old_value[index] = ui_value
            self.set_field(data_name, old_value)
        if data_name in self.properties_dict.items():
            self._slot_property_changed(data_name)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains manager object for libraries
"""

from __future__ import print_function, division, absolute_import

import os
from collections import OrderedDict

from tpDcc.core import scripts
from tpDcc.libs import qt
from tpDcc.libs.python import fileio, folder, settings, osplatform, path as path_utils
from tpDcc.libs.qt.widgets.library import items

try:
    from tpDcc.libs.python.externals.scandir import walk
except ImportError:
    from os import walk


class LibraryManager(object):
    """
    Class that manages library items registration
    """

    def __init__(self, settings=None):
        super(LibraryManager, self).__init__()

        self._library_window = None
        self._settings = settings
        self._item_classes = OrderedDict()

        self.register_item(items.LibraryFolderItem)

    def library_window(self):
        """
        Returns library window
        :return: LibraryWindow
        """

        return self._library_window

    def set_library_window(self, library_window):
        """
        Sets library window
        :param window: LibraryWindow
        """

        self._library_window = library_window

    def settings(self):
        """
        Returns LibraryManager settings
        :return: JSONSettings
        """

        return self._settings

    def set_settings(self, settings):
        """
        Sets LibraryManager settings
        :param settings: JSONSettings
        """

        self._settings = settings

    def register_item(self, cls):
        """
        Register the given item class to the given extension
        :param cls: LibraryItem
        """

        self._item_classes[cls.__name__] = cls

    def registered_items(self):
        """
        Returns all registered library item classes
        :return: list(LibraryItem)
        """

        def key(cls):
            return cls.RegisterOrder

        return sorted(self._item_classes.values(), key=key)

    def clear_registered_items(self):
        """
        Remove all registered item classes
        """

        self._item_classes = OrderedDict()

    def get_ignore_paths(self):
        """
        Returns paths that manager should ignore when creating new items
        Implements in specific class to set custom paths
        :return: list(str)
        """

        return list()

    def is_type_registered(self, data_type):
        """
        Returns whether there are classes that supports given data type
        :param data_type: str
        :return: bool
        """

        for cls in self.registered_items():
            if data_type == cls.DataType:
                return True

        return False

    def item_from_data_type(self, path, data_type):
        """
        Returns a new item instance given
        :param path: str
        :param data_type: str
        :return: LibraryItem or None
        """

        if not self.is_type_registered(data_type):
            return None

        valid_items = list()

        for cls in self.registered_items():
            if data_type == cls.DataType:
                valid_items.append(cls)

        if len(valid_items) > 1:
            qt.logger.warning(
                'Multiple data file supports data type {}:\n{}'.format(data_type, '\n'.join(valid_items)))
        valid_item = valid_items[0]

        return self.item_from_path(path, valid_item.Extension)

    def item_from_path(self, path, extension=None, **kwargs):
        """
        Return a new item instance for the given path
        :param path: str
        :param extension: str
        :param kwargs: dict
        :return: LibraryItem or None
        """

        if extension:
            full_path = path_utils.normalize_path(path + extension)
        else:
            full_path = path_utils.normalize_path(path)
        path = path_utils.normalize_path(path)
        for ignore in self.get_ignore_paths():
            if ignore in full_path:
                return None

        for cls in self.registered_items():
            if cls.match(full_path):
                lib_win = kwargs.get('library_window', self.library_window())
                kwargs['library_window'] = lib_win
                return cls(path, **kwargs)

    def items_from_paths(self, paths, **kwargs):
        """
        Return new item instances for the given paths
        :param paths: list(str)
        :param kwargs: dict
        :return: Iterable(LibraryItem)
        """

        for path in paths:
            item = self.item_from_path(path, **kwargs)
            if item:
                yield item

    def items_from_urls(self, urls, **kwargs):
        """
        Return new item instances for the given QUrl objects
        :param urls: list(QUrl)
        :param kwargs: dict
        :return: list(LibraryItem)
        """

        items = list()
        for path in self.paths_from_urls(urls):
            item = self.item_from_path(path, **kwargs)
            if item:
                items.append(item)
            else:
                msg = 'Cannot find the item for path "{}"'.format(path)
                qt.logger.warning(msg)

        return items

    def paths_from_urls(self, urls):
        """
        Returns the local file paths from the given QUrls
        :param urls: list(QUrl)
        :return: Iterable(str)
        """

        for url in urls:
            path = url.toLocalFile()

            if osplatform.is_windows():
                if path.startswith('/'):
                    path = path[1:]

            yield path

    def find_items(self, path, depth=3, **kwargs):
        """
        Find and create items by walking the given path
        :param path: str
        :param depth: int
        :param kwargs: dict
        :return: Iterable(LibraryItem)
        """

        path = path_utils.normalize_path(path)
        max_depth = depth
        start_depth = path.count(os.path.sep)

        for root, dirs, files in walk(path, followlinks=True):
            files.extend(dirs)
            for filename in files:
                remove = False
                path = os.path.join(root, filename)
                item = self.item_from_path(path, **kwargs)
                if item:
                    yield item
                    if not item.EnableNestedItems:
                        remove = True
                if remove and filename in dirs:
                    dirs.remove(filename)

            if depth == 1:
                break

            # Stop walking the directory if the maximum depth has been reached
            current_depth = root.count(os.path.sep)
            if (current_depth - start_depth) >= max_depth:
                del dirs[:]

    def find_items_in_folders(self, folders, depth=3, **kwargs):
        """
        Find and create new item instances by walking the given paths
        :param folders: list(str)
        :param depth: int
        :param kwargs: dict
        :return: Iterable(LibraryItem)
        """

        for folder in folders:
            for item in self.find_items(folder, depth=depth, **kwargs):
                yield item


class LibraryDataFolder(fileio.FileManager, object):
    def __init__(self, name, file_path, data_path=None):
        super(LibraryDataFolder, self).__init__(file_path=file_path)

        self.settings = None
        new_path = path_utils.join_path(file_path, name)
        self.file_path = path_utils.get_dirname(new_path)
        self.name = path_utils.get_basename(new_path)
        self.data_type = None

        # if data_path:
        #     tp.data_manager.add_directory(data_path)

        # Check if the given file path is a folder or not (and create folder if necessary)
        test_path = path_utils.join_path(self.file_path, self.name)
        is_folder = path_utils.is_dir(test_path)
        if is_folder:
            self.folder_path = test_path
        else:
            self._create_folder()

    def get_manager(self):
        """
        Returns LibraryManager that should used to create data folder instance
        :return: LibraryManager
        """

        raise NotImplementedError('get_manager not implemented in LibraryDataFolder')

    def get_data_type(self):
        """
        Returns the type of data stored in this data folder
        :return: str
        """

        if self.settings:
            self.settings.reload()
        else:
            self._load_folder()

        return self.settings.get('data_type')

    def set_data_type(self, data_type):
        """
        Set the type of data stored in this data folder
        :param data_type: str
        """

        if not self.settings:
            self._load_folder()

        self.data_type = data_type
        self.settings.set('data_type', str(data_type))

    def get_sub_folder(self, name=None):
        """
        Returns sub folder of this data folder
        :param name: str, name of the sub folder
        :return: str
        """

        if name:
            sub_folder = name
        else:
            if not self.settings:
                self._load_folder()
            sub_folder = self.settings.get('sub_folder')

        if self.folder_path:
            if not path_utils.is_dir(path_utils.join_path(self.folder_path, '.sub/{}'.format(sub_folder))):
                return

        return folder

    def set_sub_folder(self, name):
        """
        Sets the sub folder of this data folder
        :param name: str, name of the sub folder
        """

        if not name:
            return

        if not self.settings:
            self._load_folder()

        self.settings.set('sub_folder', name)
        sub_folder = path_utils.join_path(self.folder_path, '.sub/{}'.format(name))
        folder.create_folder(sub_folder)
        self.settings.set('data_type', str(self.data_type))

    def get_current_sub_folder(self):
        """
        Return current sub folder of data folder
        :return: str
        """

        sub_folder = self.get_sub_folder()
        return sub_folder

    def rename(self, new_name):
        """
        Rename this data folder
        :param new_name: str
        :return: str, new folder name
        """

        base_name = path_utils.get_basename(new_name)
        data_inst = self.get_folder_data_instance()
        data_inst.rename(base_name)
        renamed_folder = folder.rename_folder(self.folder_path, new_name)
        if not renamed_folder:
            return

        self.folder_path = renamed_folder
        self._set_settings_path(renamed_folder)
        self._set_name(base_name)

        return self.folder_path

    def delete(self):
        """
        Removes data folder from disk
        """

        folder_name = path_utils.get_basename(self.folder_path)
        data_directory = path_utils.get_dirname(self.folder_path)

        folder.delete_folder(folder_name, data_directory)

    def get_folder_data_instance(self):
        """
        Returns the data instance for this data
        Depending on the data type this folder contains a differnet object is created
        :return: variant
        """

        if not self.settings:
            self._load_folder()

        if not self.name:
            return

        data_type = self.settings.get('data_type')
        if not data_type:
            data_type = self.data_type

        if data_type is None:
            test_file = os.path.join(self.folder_path, '{}.{}'.format(self.name, scripts.ScriptExtensions.Python))
            if path_utils.is_file(test_file):
                data_type = scripts.ScriptTypes.Python
                self.settings.set('data_type', data_type)
        if not data_type:
            qt.logger.warning(
                'Impossible to instantiate Data Folder because given Data Type: {} is not valid!'.format(data_type))
            return

        data_manager = self.get_manager()
        if not data_manager:
            qt.logger.warning('Impossible to instantiate Data Folder because LibraryManager is not defined!')
            return

        data_inst = data_manager.get_type_instance(data_type)
        if data_inst:
            data_inst.set_directory(self.folder_path)
            data_inst.set_name(self.name)

        return data_inst

    def _load_folder(self):
        """
        Internal function that load settings with the current contents of the data folder
        """

        self._set_default_settings()

    def _set_name(self, name):
        """
        Internal function that stores given name into the data folder settings file
        :param name: str, new name
        """

        if not self.settings:
            self._load_folder()

        self.name = name
        self.settings.set('name', self.name)

    def _set_settings_path(self, folder_path):
        """
        Internal function that stores the given path into the data folder settings file
        :param folder_path: str, folder path
        """

        if not self.settings:
            self._load_folder()

        self.settings.set_directory(folder_path, 'data.json')

    def _set_default_settings(self):
        """
        Internal function that initializes the settings of the data folder
        """

        self.settings = settings.JSONSettings()
        self._set_settings_path(self.folder_path)
        self.settings.set('name', self.name)
        self.data_type = self.settings.get('data_type')

    def _create_folder(self):
        """
        Internal function that creates data folder
        """

        data_path = folder.create_folder(self.name, self.file_path)
        self.folder_path = data_path

        self._set_default_settings()



order = [
    'tpDcc.libs.qt.widgets.layouts',
    'tpDcc.libs.qt.widgets.graphicseffects',
    'tpDcc.libs.qt.widgets.buttons',
    'tpDcc.libs.qt.widgets.group',
    'tpDcc.libs.qt.widgets.label',
    'tpDcc.libs.qt.widgets.avatar',
    'tpDcc.libs.qt.widgets.lineedit',
    'tpDcc.libs.qt.widgets.dividers',
    'tpDcc.libs.qt.widgets.gif',
    'tpDcc.libs.qt.widgets.directory',
    'tpDcc.libs.qt.widgets.grid',
    'tpDcc.libs.qt.widgets.search',
    'tpDcc.libs.qt.widgets.treewidgets',
    'tpDcc.libs.qt.widgets.history',
    'tpDcc.libs.qt.widgets.tabbars',
    'tpDcc.libs.qt.widgets.tabs',
    'tpDcc.libs.qt.widgets.overlay',
    'tpDcc.libs.qt.widgets.dpi',
    'tpDcc.libs.qt.widgets.stack',
    'tpDcc.libs.qt.widgets.loading',
    'tpDcc.libs.qt.widgets.message',
    'tpDcc.libs.qt.widgets.combobox',
    'tpDcc.libs.qt.widgets.spinbox',
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpDcc.libs.qt
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import inspect
import logging

from Qt.QtCore import *
from Qt.QtWidgets import *

main = __import__('__main__')


def init(do_reload=False, dev=False):
    """
    Initializes module
    :param do_reload: bool, Whether to reload modules or not
    :param dev: bool, Whether artellapipe is initialized in dev mode or not
    """

    from tpDcc.libs.qt import register
    from tpDcc.libs.python import importer

    logger = create_logger(dev=dev)

    class tpQtLib(importer.Importer, object):
        def __init__(self, *args, **kwargs):
            super(tpQtLib, self).__init__(module_name='tpDcc.libs.qt', *args, **kwargs)

        def get_module_path(self):
            """
            Returns path where tpQtLib module is stored
            :return: str
            """

            try:
                mod_dir = os.path.dirname(inspect.getframeinfo(inspect.currentframe()).filename)
            except Exception:
                try:
                    mod_dir = os.path.dirname(__file__)
                except Exception:
                    try:
                        import tpDcc.libs.qt
                        mod_dir = tpDcc.libs.qt.__path__[0]
                    except Exception:
                        return None

            return mod_dir

        def update_paths(self):
            """
            Adds path to system paths at startup
            """

            paths_to_update = [self.externals_path()]

            for p in paths_to_update:
                if os.path.exists(p) and p not in sys.path:
                    sys.path.append(p)

        def externals_path(self):
            """
            Returns the paths where tpPyUtils externals packages are stored
            :return: str
            """

            return os.path.join(self.get_module_path(), 'externals')

    def init_dcc(do_reload=False):
        """
        Checks DCC we are working on an initializes proper variables
        """

        if 'cmds' in main.__dict__:
            from tpDcc.dccs.maya import loader
            loader.init_ui(do_reload=do_reload)
        elif 'MaxPlus' in main.__dict__:
            from tpDcc.dccs.max import loader
            loader.init_ui(do_reload=do_reload)
        elif 'hou' in main.__dict__:
            from tpDcc.dccs.houdini import loader
            loader.init_ui(do_reload=do_reload)
        elif 'nuke' in main.__dict__:
            from tpDcc.dccs.nuke import loader
        else:
            global Dcc
            from tpDcc.core import dcc
            Dcc = dcc.UnknownDCC
            logger.warning('No DCC found, using abstract one!')

        from tpDcc.managers import callbacks
        callbacks.CallbacksManager.initialize()

    app = QApplication.instance()
    if not app:
        app = QApplication(sys.argv)

    qt_importer = importer.init_importer(importer_class=tpQtLib, do_reload=False)
    qt_importer.update_paths()

    register.register_class('logger', logger)

    qt_importer.import_modules(skip_modules=['tpDcc.libs.qt.externals'])
    qt_importer.import_packages(only_packages=True, skip_modules=['tpDcc.libs.qt.externals'],
                                order=['tpDcc.libs.qt.core', 'tpDcc.libs.qt.widgets'])
    if do_reload:
        qt_importer.reload_all()

    init_dcc(do_reload=do_reload)

    register_resources()


def create_logger(dev=False):
    """
    Returns logger of current module
    """

    logging.config.fileConfig(get_logging_config(), disable_existing_loggers=False)
    logger = logging.getLogger('tpDcc-libs-qt')
    if dev:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    return logger


def create_logger_directory():
    """
    Creates artellapipe logger directory
    """

    tppyutils_logger_dir = os.path.normpath(os.path.join(os.path.expanduser('~'), 'tpDcc', 'logs'))
    if not os.path.isdir(tppyutils_logger_dir):
        os.makedirs(tppyutils_logger_dir)


def get_logging_config():
    """
    Returns logging configuration file path
    :return: str
    """

    create_logger_directory()

    return os.path.normpath(os.path.join(os.path.dirname(__file__), '__logging__.ini'))


def register_resources():
    """
    Registers tpDcc.libs.qt resources path
    """

    import tpDcc

    resources_manager = tpDcc.ResourcesMgr()
    resources_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources')
    resources_manager.register_resource(resources_path, key='tpDcc-libs-qt')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Register module for tpDcc-libs-qt
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"


def register_class(cls_name, cls, is_unique=False):
    """
    This function registers given class into tpRigToolkit module
    :param cls_name: str, name of the class we want to register
    :param cls: class, class we want to register
    :param is_unique: bool, Whether if the class should be updated if new class is registered with the same name
    """

    import tpDcc.libs.qt

    if is_unique:
        if cls_name in tpDcc.libs.qt.__dict__:
            setattr(tpDcc.libs.qt.__dict__, cls_name, getattr(tpDcc.libs.qt.__dict__, cls_name))
    else:
        tpDcc.libs.qt.__dict__[cls_name] = cls



# This file is part of the PySide project.
#
# Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).
# Copyright (C) 2010 Riverbank Computing Limited.
#
# Contact: PySide team <pyside@openbossa.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301 USA

import re


def as_string(obj, encode=True):
    if isinstance(obj, basestring):
        s = '"' + _escape(obj.encode('UTF-8')) + '"'
        return s

    return str(obj)


_esc_regex = re.compile(r"(\"|\'|\\)")

def _escape(text):
    # This escapes any escaped single or double quote or backslash.
    x = _esc_regex.sub(r"\\\1", text)

    # This replaces any '\n' with an escaped version and a real line break.
    return re.sub(r'\n', r'\\n"\n"', x)



# This file is part of the PySide project.
#
# Copyright (C) 2009-2011 Nokia Corporation and/or its subsidiary(-ies).
# Copyright (C) 2010 Riverbank Computing Limited.
# Copyright (C) 2009 Torsten Marek
#
# Contact: PySide team <pyside@openbossa.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301 USA


import logging

try:
    set()
except NameError:
    from sets import Set as set

from pysideuic.Compiler.indenter import write_code
from pysideuic.Compiler.qtproxies import QtGui, Literal, strict_getattr


logger = logging.getLogger(__name__)
DEBUG = logger.debug


class _QtGuiWrapper(object):
    def search(clsname):
        try:
            return strict_getattr(QtGui, clsname)
        except AttributeError:
            return None

    search = staticmethod(search)


class _ModuleWrapper(object):
    def __init__(self, name, classes):
        if "." in name:
            idx = name.rfind(".")
            self._package = name[:idx]
            self._module = name[idx + 1:]
        else:
            self._package = None
            self._module = name

        self._classes = set(classes)
        self._used = False

    def search(self, cls):
        if cls in self._classes:
            self._used = True
            return type(cls, (QtGui.QWidget,), {"module": self._module})
        else:
            return None

    def _writeImportCode(self):
        if self._used:
            if self._package is None:
                write_code("import %s" % self._module)
            else:
                write_code("from %s import %s" % (self._package, self._module))


class _CustomWidgetLoader(object):
    def __init__(self):
        self._widgets = {}
        self._usedWidgets = set()

    def addCustomWidget(self, widgetClass, baseClass, module):
        assert widgetClass not in self._widgets
        self._widgets[widgetClass] = (baseClass, module)


    def _resolveBaseclass(self, baseClass):
        try:
            for x in range(0, 10):
                try: return strict_getattr(QtGui, baseClass)
                except AttributeError: pass

                baseClass = self._widgets[baseClass][0]
            else:
                raise ValueError("baseclass resolve took too long, check custom widgets")

        except KeyError:
            raise ValueError("unknown baseclass %s" % baseClass)


    def search(self, cls):
        try:
            self._usedWidgets.add(cls)
            baseClass = self._resolveBaseclass(self._widgets[cls][0])
            DEBUG("resolved baseclass of %s: %s" % (cls, baseClass))

            return type(cls, (baseClass,),
                        {"module" : ""})

        except KeyError:
            return None

    def _writeImportCode(self):
        imports = {}
        for widget in self._usedWidgets:
            _, module = self._widgets[widget]
            imports.setdefault(module, []).append(widget)

        for module, classes in imports.items():
            write_code("from %s import %s" % (module, ", ".join(classes)))


class CompilerCreatorPolicy(object):
    def __init__(self):
        self._modules = []

    def createQtGuiWrapper(self):
        return _QtGuiWrapper

    def createModuleWrapper(self, name, classes):
        mw = _ModuleWrapper(name, classes)
        self._modules.append(mw)
        return mw

    def createCustomWidgetLoader(self):
        cw = _CustomWidgetLoader()
        self._modules.append(cw)
        return cw

    def instantiate(self, clsObject, objectname, ctor_args, is_attribute=True, no_instantiation=False):
        return clsObject(objectname, is_attribute, ctor_args, no_instantiation)

    def invoke(self, rname, method, args):
        return method(rname, *args)

    def getSlot(self, object, slotname):
        return Literal("%s.%s" % (object, slotname))

    def _writeOutImports(self):
        for module in self._modules:
            module._writeImportCode()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains delegates that can be used to customize how data is renderer in list views
"""

from __future__ import print_function, division, absolute_import

import os

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *


class BaseListViewDelegate(QStyledItemDelegate, object):
    """
    Base delegate for list view widgets
    """

    _ICON_MARGIN = 4

    # region Override Functions
    def paint(self, painter, option, index):

        painter.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)
        model = index.model()
        view = self.parent()

        if view.hasFocus() and option.state & QStyle.State_MouseOver:
            painter.setPen(Qt.NoPen)
            painter.setBrush(Qt.gray)
            painter.drawRoundedRect(option.rect.adjusted(1, 1, -1, -1), self._ICON_MARGIN, self._ICON_MARGIN)

        pixmap = model.data(index, Qt.DecorationRole).pixmap(view.iconSize())
        pm_rect = QRect(option.rect.topLeft() + QPoint(self._ICON_MARGIN + 1, self._ICON_MARGIN + 1),
                        view.iconSize() - QSize(self._ICON_MARGIN * 2, self._ICON_MARGIN * 2))
        painter.drawPixmap(pm_rect, pixmap)
        if option.state & QStyle.State_Selected:
            painter.setPen(QPen(Qt.red, 1.0, Qt.SolidLine, Qt.SquareCap, Qt.RoundJoin))
            painter.setBrush(Qt.NoBrush)
            painter.drawRect(option.rect.adjusted(2, 2, -2, -2))

        font = view.font()
        fm = QFontMetrics(font)
        text = os.path.splitext(os.path.basename(model.data(index, Qt.DisplayRole)))[0]
        text = fm.elidedText(text, Qt.ElideRight, view.iconSize().width() - 4)
        text_opt = QTextOption()
        text_opt.setAlignment(Qt.AlignHCenter)
        txt_rect = QRectF(
            QPointF(pm_rect.bottomLeft() + QPoint(0, 1)), QPointF(option.rect.bottomRight() - QPoint(4, 3)))

        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor(22, 22, 22, 220))
        painter.drawRoundedRect(txt_rect.adjusted(-2, -2, 2, 2), 2, 2)
        painter.restore()
        painter.setPen(self.parent().palette().color(QPalette.WindowText))
        painter.drawText(txt_rect, text, text_opt)

        font.setPointSize(8)
        fm = QFontMetrics(font)
        item = model.itemFromIndex(index)
        size_text = '%d x %d' % (item.size.width(), item.size.height())
        size_rect = fm.boundingRect(option.rect, Qt.AlignLeft | Qt.AlignTop, size_text)
        size_rect.translate(4, 4)

        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor(22, 22, 22, 220))
        painter.drawRoundedRect(size_rect.adjusted(-2, -2, 2, 2), 2, 2)
        painter.restore()
        painter.setFont(font)
        painter.drawText(size_rect, size_text)

    def sizeHint(self, option, index):
        view = self.parent()
        return view.iconSize() + QSize(2, 14)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility module that contains functions related with formatters
"""

from __future__ import print_function, division, absolute_import

from singledispatch import singledispatch


def apply_formatter(formatter, *args, **kwargs):
    """
    Used by QAbstractModel data method
    Configures a formatter for one field, apply the formatter with the new index data
    :param formatter: formatter. If can be None/dict/callback or just any type of value
    :param args:
    :param kwargs:
    :return:
    """

    if formatter is None:
        return args[0]
    elif isinstance(formatter, dict):
        return formatter.get(args[0], None)
    elif callable(formatter):
        return formatter(*args, **kwargs)

    return formatter


def overflow_format(num, overflow):
    """
    Returns string of the given integer. If the integer is large than given overflow, '{overflow}+' is returned
    :param num: int
    :param overflow: int
    :return: str
    """

    if not isinstance(num, int):
        raise ValueError('Input argument "num" should be int type, but get "{}"'.format(type(num)))
    if not isinstance(overflow, int):
        raise ValueError('Input argument "overflow" should be int type, but get "{}"'.format(type(num)))

    return str(num) if num <= overflow else '{}+'.format(overflow)


@singledispatch
def display_formatter(input_other_type):
    """
    Used by QAbstractItemModel data method for Qt.DisplayRole
    Format any input value to a string
    :param input_other_type:
    :return: str
    """

    return str(input_other_type)


@display_formatter.register(type(None))
def _(input_none):
    return '--'


@display_formatter.register(int)
def _(input_int):
    return str(input_int)


@display_formatter.register(float)
def _(input_float):
    return '{:.2f}'.format(round(input_float, 2))


@display_formatter.register(dict)
def _(input_dict):
    if 'name' in input_dict.keys():
        return display_formatter(input_dict.get('name'))
    elif 'code' in input_dict.keys():
        return display_formatter(input_dict.get('code'))
    return str(input_dict)


@display_formatter.register(list)
def _(input_list):
    result = list()
    for i in input_list:
        result.append(display_formatter(i))
    return '.'.join(result)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementations for different types of progress bars
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *


class BaseProgressBar(QProgressBar, object):

    ERROR_STATUS = 'error'
    NORMAL_STATUS = 'primary'
    SUCCESS_STATUS = 'success'

    def __init__(self, parent=None):
        super(BaseProgressBar, self).__init__(parent)

        self.setAlignment(Qt.AlignCenter)
        self._status = BaseProgressBar.NORMAL_STATUS

    def _get_status(self):
        return self._status

    def _set_status(self, value):
        self._status = value
        self.style().polish(self)

    theme_status = Property(str, _get_status, _set_status)

    def normal(self):
        self.theme_status = BaseProgressBar.NORMAL_STATUS

    def success(self):
        self.theme_status = BaseProgressBar.SUCCESS_STATUS

    def error(self):
        self.theme_status = BaseProgressBar.ERROR_STATUS

    def auto_color(self):
        self.valueChanged.connect(self._on_update_color)

        return self

    def _on_update_color(self, value):
        if value >= self.maximum():
            self.theme_status = BaseProgressBar.SUCCESS_STATUS
        else:
            self.theme_status = BaseProgressBar.NORMAL_STATUS


class FrameProgressBar(QFrame, object):
    def __init__(self, *args):
        super(FrameProgressBar, self).__init__(*args)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._label = QLabel('', self)
        self._label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        layout.addWidget(self._label)

        self._progress_bar = QProgressBar(self)
        self._progress_bar.setFormat('')
        self._progress_bar.setRange(0, 100)
        self._progress_bar.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        layout.addWidget(self._progress_bar)

    def reset(self):
        """
        Reset progress bar
        """

        self._progress_bar.reset()

    def set_text(self, text):
        """
        Set the text of the progress bar
        :param text: str
        """

        self._label.setText(text)

    def set_value(self, value):
        """
        Set the value of the progress bar
        :param value: int or float
        """

        self._progress_bar.setValue(value)

    def set_range(self, min_, max_):
        """
        Set the range of the progress bar
        :param min_: int
        :param max_: int
        """

        self._progress_bar.setRange(min_, max_)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different graphic scenes
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.python import decorators


class BaseScene(QGraphicsScene, object):
    """
    Base scene for graphics scenes that add support for z index
    """

    scene_changed = Signal()

    _DEFAULT_SCENE_WIDTH = 100
    _DEFAULT_SCENE_HEIGHT = 200

    def __init__(self, auto_z=False, parent=None):
        super(BaseScene, self).__init__(parent=parent)

        self._auto_z = auto_z
        self._z_index = 0
        self._root = None
        self.setParent(parent)

    def get_root(self):
        """
        Returns root scene object
        :return: QSceneObject
        """

        return self._root

    def set_root(self, value):
        """
        Set the root scene object
        :param value: QSceneObject
        """

        self._root = value
        self._root.setParent(self)

    def get_auto_z(self):
        """
        Returns whether auto Z is enabled or not
        :return: bool
        """

        return self._auto_z

    def set_auto_z(self, auto_z):
        """
        Sets whether auto z is enabled or not
        :param auto_z: bool
        """

        self._auto_z = auto_z

    def set_default_size(self):
        """
        Resets the view to its default values
        """

        return self.set_size(self._DEFAULT_SCENE_WIDTH, self._DEFAULT_SCENE_HEIGHT)

    def set_size(self, width, height):
        """
        Sets the size of the scene
        :param width: int, new width of the scene
        :param height: int, new height of the scene
        """

        self.setSceneRect(0, 0, width, height)

    def get_items_by_z_value_order(self, classes_tuple=None, rect=QRectF()):
        """
        Returns the items of the passed class
        :param classes_tuple:
        :param rect:
        :return:
        """

        def cmp_zvalue(x, y):
            x_val = x.zValue()
            y_val = y.zValue()
            if x_val > y_val:
                return 1
            elif x_val < y_val:
                return -1
            else:
                return 0

        if classes_tuple is not None:
            items = filter(lambda x: isinstance(x, classes_tuple), rect.isValid() and self.items(rect) or self.items())
        else:
            items = self.items()
        if items:
            items.sort(key=lambda x: x.zValue())
        return items

    def get_top_item(self):
        """
        Get top Z value item of the scene
        :return:
        """

        items = self.get_items_by_z_value_order()
        if items:
            return items[-1]
        else:
            return None

    def primary_view(self):
        """
        Returns the first view of the scene
        """

        views = self.views()
        if views:
            return views[0]

    def window(self):
        """
        Returns the view window
        """

        view = self.primary_view()
        if view:
            return view.window()
        else:
            return ''

    def clear(self):
        """
        Reset default Z index on clear
        """

        super(BaseScene, self).clear()
        self._z_index = 0

    def addItem(self, item):
        super(BaseScene, self).addItem(item)

        if self._auto_z:
            self._set_z_value(item)

    def _set_z_value(self, control):

        """
        Sets a proper Z idnex for the control
        """

        control.setZValue(self._z_index)
        self._z_index += 1


class GridColors(object):
    """
    Class that stores predefined colors for grids
    """

    BaseColor = QColor(60, 60, 60, 100)
    DarkerColor = QColor(20, 20, 20, 100)


class BackgroundImageScene(BaseScene, object):
    """
    Scene with image background drawing support
    """

    def __init__(self, parent=None):
        super(BackgroundImageScene, self).__init__(parent=parent)

        self._image_path = ''
        self._use_bg_image = True
        self._pixmap = QPixmap()

    @decorators.accepts(QPixmap)
    def get_pixmap(self):
        return self._pixmap

    @decorators.returns(QPixmap)
    def set_pixmap(self, value):
        self._pixmap = value

    @decorators.returns(bool)
    def get_use_bg_image(self):
        return self._use_bg_image

    @decorators.accepts(bool)
    def set_use_bg_image(self, value):
        self._use_bg_image = value
        self.update()

    pixmap = property(get_pixmap, set_pixmap)
    use_bg_image = property(get_use_bg_image, set_use_bg_image)

    def set_background_pixmap(self, pixmap_path):
        """
        Sets the background of the scene
        :param pixmap_path: str, path to the pixmap
        """

        pixmap = QPixmap(pixmap_path)
        if not pixmap.isNull():
            self._image_path = pixmap_path
            self._pixmap = pixmap
        else:
            self._image_path = ''
            self._pixmap = QPixmap()
        self.use_bg_image = True

    def clear_background_image(self):
        """
        Clears the background image
        """

        self._pixmap = QPixmap()
        self._image_path = ''
        self.use_bg_image = False

    def drawBackground(self, painter, rect):
        """
        Override draw background method to write out image as background
        """

        painter.setRenderHint(QPainter.Antialiasing)
        painter.fillRect(rect, self._color)
        painter.setPen(QPen(Qt.black, 0.5, Qt.DashLine))
        painter.setBrush(QColor(67, 255, 163))
        if self._use_bg_image and not self._pixmap.isNull():
            painter.drawPixmap(QPointF(0, 0), self._pixmap)


class GridScene(BaseScene, object):
    """
    Scene with grid background drawing support
    """

    def __init__(self,
                 grid_main_spacing=8,
                 grid_secondary_spacing=8,
                 grid_main_pen_color=GridColors.BaseColor,
                 grid_secondary_pen_color=GridColors.DarkerColor,
                 grid_main_width=1,
                 grid_secondary_width=1,
                 grid_main_style=Qt.SolidLine,
                 grid_secondary_style=Qt.SolidLine,
                 draw_main_grid=True,
                 draw_secondary_grid=True,
                 fit_grid=False,
                 fit_grid_main_divisions=10,
                 fit_grid_secondary_divisions=4,
                 parent=None):
        super(GridScene, self).__init__(parent=parent)

        self._grid_main_pen = QPen()
        self._grid_secondary_pen = QPen()
        self._grid_main_spacing = grid_main_spacing
        self._grid_secondary_spacing = grid_secondary_spacing
        self._grid_main_pen.setColor(grid_main_pen_color)
        self._grid_secondary_pen.setColor(grid_secondary_pen_color)
        self._grid_main_pen.setWidth(grid_main_width)
        self._grid_secondary_pen.setWidth(grid_secondary_width)
        self._grid_main_style = grid_main_style
        self._grid_secondary_style = grid_secondary_style
        self._draw_main_grid = draw_main_grid
        self._draw_secondary_grid = draw_secondary_grid
        self._fit_grid = fit_grid
        self._fit_grid_main_divisions = fit_grid_main_divisions
        self._fit_grid_secondary_divisions = fit_grid_secondary_divisions

        self._grid_main_pen.setStyle(self._grid_main_style)
        self._grid_secondary_pen.setStyle(self._grid_secondary_style)
        self._fit_grid_draw = False

    @decorators.accepts(int)
    def set_grid_main_spacing(self, value):
        """
        Set the size of the main grid (define the gap between grid lines)
        """

        self._grid_main_spacing = value

    @decorators.accepts(int)
    def set_grid_secondary_spacing(self, value):
        """
        Set the size of the secondary grid
        """

        self._grid_secondary_spacing = value

    @decorators.returns(int)
    def get_grid_main_spacing(self):
        """
        Returns the size of the main grid
        :return: int, size of the main grid
        """

        return self._grid_main_spacing

    @decorators.returns(int)
    def get_grid_secondary_spacing(self):
        """
        Returns the size of the secondary grid
        :return: int, size of the secondary grid
        """

        return self._grid_secondary_spacing

    @decorators.accepts(QColor)
    def set_main_pen_color(self, value):
        self._grid_main_pen.setColor(value)

    @decorators.returns(QColor)
    def get_main_pen_color(self):
        return self._grid_main_pen.color()

    @decorators.accepts(QColor)
    def set_secondary_pen_color(self, value):
        self._grid_secondary_pen.setColor(value)

    @decorators.returns(QColor)
    def get_secondary_pen_color(self):
        return self._grid_secondary_pen.color()

    @decorators.accepts(float)
    def set_grid_width(self, value):
        self._grid_main_pen.setWidth(value)
        self._grid_secondary_pen.setWidth(value)

    @decorators.returns(float)
    def get_grid_width(self):
        return self._grid_main_pen.width()

    @decorators.returns(bool)
    def get_draw_main_grid(self):
        return self._draw_main_grid

    @decorators.accepts(bool)
    def set_draw_main_grid(self, draw):
        self._draw_main_grid = draw

    @decorators.returns(bool)
    def get_draw_secondary_grid(self):
        return self._draw_secondary_grid

    @decorators.accepts(bool)
    def set_draw_secondary_grid(self, draw):
        self._draw_secondary_grid = draw

    @decorators.returns(bool)
    def get_fit_grid(self):
        return self._fit_grid

    @decorators.accepts(bool)
    def set_fit_grid(self, fit):
        self._fit_grid = fit

    grid_main_spacing = property(get_grid_main_spacing, set_grid_main_spacing)
    grid_secondary_spacing = property(get_grid_secondary_spacing, set_grid_secondary_spacing)
    grid_main_pen_color = property(get_main_pen_color, set_main_pen_color)
    grid_secondary_pen_color = property(get_secondary_pen_color, set_secondary_pen_color)
    grid_width = property(get_grid_width, set_grid_width)
    draw_main_grid = property(get_draw_main_grid, set_draw_main_grid)
    draw_secondary_grid = property(get_draw_secondary_grid, set_draw_secondary_grid)
    fit_grid = property(get_fit_grid, set_fit_grid)

    # region Override Functions
    def drawBackground(self, painter, rect):
        """
        Draw grid background for the graph scene
        """

        scene_rect = self.sceneRect()
        gradient = QColor(65, 65, 65)
        painter.fillRect(rect.intersected(scene_rect), QBrush(gradient))
        painter.setPen(QPen())
        painter.drawRect(scene_rect)

        if len(self.views()) <= 0:
            tp.logger.error('Scene has not view associated to it!')
            return
        if not self.views()[0]:
            tp.logger.error('View {0} is not valid!'.format(self.views()[0]))
            return
        if hasattr(self.views()[0], 'is_grid_visible'):
            if self.views()[0].is_grid_visible:
                return

        if self._fit_grid:
            if not self._fit_grid_draw:
                if self._draw_main_grid:
                    pos = float(self.width() / self._fit_grid_main_divisions)
                    for div in range(self._fit_grid_main_divisions):
                        if div == 0:
                            continue
                        line = QGraphicsLineItem(0, pos * div, self.width(), pos * div)
                        line.setZValue(-1)
                        line.setPen(self._grid_main_pen)
                        self.addItem(line)
                        line = QGraphicsLineItem(pos * div, 0, pos * div, self.width())
                        line.setZValue(-1)
                        line.setPen(self._grid_main_pen)
                        self.addItem(line)
                    rect = QGraphicsRectItem(self.sceneRect())
                    rect.setZValue(-1)
                    self.addItem(rect)

                if self._draw_secondary_grid:
                    pos = float(self.width() / self._fit_grid_secondary_divisions)
                    for div in range(self._fit_grid_secondary_divisions):
                        if div == 0:
                            continue
                        line = QGraphicsLineItem(0, pos * div, self.width(), pos * div)
                        line.setZValue(-1)
                        line.setPen(self._grid_secondary_pen)
                        self.addItem(line)
                        line = QGraphicsLineItem(pos * div, 0, pos * div, self.width())
                        line.setZValue(-1)
                        line.setPen(self._grid_secondary_pen)
                        self.addItem(line)
                    rect = QGraphicsRectItem(self.sceneRect())
                    rect.setZValue(-1)
                    self.addItem(rect)
                self._fit_grid_draw = True
        else:

            if self._draw_main_grid:
                left = int(self.sceneRect().left()) - (int(self.sceneRect().left()) % self._grid_main_spacing)
                top = int(self.sceneRect().top()) - (int(self.sceneRect().top()) % self._grid_main_spacing)

                painter.setPen(self._grid_main_pen)

                # draw grid vertical lines
                for x in range(left, int(self.sceneRect().right()), self.grid_main_spacing):
                    painter.drawLine(x, self.sceneRect().top(), x, self.sceneRect().bottom())

                # draw grid horizontal lines
                for y in range(top, int(self.sceneRect().bottom()), self.grid_main_spacing):
                    painter.drawLine(self.sceneRect().left(), y, self.sceneRect().right(), y)

            if self._draw_secondary_grid:
                left = int(self.sceneRect().left()) - (int(self.sceneRect().left()) % self._grid_secondary_spacing)
                top = int(self.sceneRect().top()) - (int(self.sceneRect().top()) % self._grid_secondary_spacing)

                painter.setPen(self._grid_secondary_pen)

                # draw grid vertical lines
                for x in range(left, int(self.sceneRect().right()), self.grid_secondary_spacing * 10):
                    painter.drawLine(x, self.sceneRect().top(), x, self.sceneRect().bottom())

                # draw grid horizontal lines
                for y in range(top, int(self.sceneRect().bottom()), self.grid_secondary_spacing * 10):
                    painter.drawLine(self.sceneRect().left(), y, self.sceneRect().right(), y)
    # endregion


class DropEditableScene(GridScene, object):
    def __init__(self, parent=None):
        super(DropEditableScene, self).__init__(parent=parent)

        self._editable = True

    @decorators.returns(bool)
    def get_editable(self):
        return self._editable

    @decorators.accepts(bool)
    def set_editable(self, value):
        self._editable = value
        self.update()

    def get_bounding_rect(self, margin=0):
        """
        Return scene content bounding box with specified margin
        """

        if self._editable:
            return self.sceneRect()
        scene_rect = self.itemsBoundingRect()
        if not margin:
            return scene_rect
        scene_rect.setX(scene_rect.x() - margin)
        scene_rect.setY(scene_rect.y() - margin)
        scene_rect.setWidth(scene_rect.width() + margin)
        scene_rect.setHeight(scene_rect.height() + margin)
        return scene_rect
